########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import io
import pathlib
import tempfile
import unittest
import unittest.mock

import aioxmpp.disco as disco
import aioxmpp.service as service
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs
import aioxmpp.xml

from aioxmpp.utils import namespaces

import aioxmpp.entitycaps.service as entitycaps_service

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)


TEST_FROM = structs.JID.fromstr("foo@bar.example/r1")


_src = io.BytesIO(b"""\
<?xml version="1.0" ?><query node="http://tkabber.jabber.ru/#+0mnUAF1ozCEc37cm\
dPPsYbsfhg=" xmlns="http://jabber.org/protocol/disco#info"><identity category=\
"client" name="Tkabber" type="pc"/><feature var="games:board"/><feature var="h\
ttp://jabber.org/protocol/activity"/><feature var="http://jabber.org/protocol/\
activity+notify"/><feature var="http://jabber.org/protocol/bytestreams"/><feat\
ure var="http://jabber.org/protocol/chatstates"/><feature var="http://jabber.o\
rg/protocol/commands"/><feature var="http://jabber.org/protocol/disco#info"/><\
feature var="http://jabber.org/protocol/disco#items"/><feature var="http://jab\
ber.org/protocol/geoloc"/><feature var="http://jabber.org/protocol/geoloc+noti\
fy"/><feature var="http://jabber.org/protocol/ibb"/><feature var="http://jabbe\
r.org/protocol/iqibb"/><feature var="http://jabber.org/protocol/mood"/><featur\
e var="http://jabber.org/protocol/mood+notify"/><feature var="http://jabber.or\
g/protocol/rosterx"/><feature var="http://jabber.org/protocol/si"/><feature va\
r="http://jabber.org/protocol/si/profile/file-transfer"/><feature var="http://\
jabber.org/protocol/tune"/><feature var="jabber:iq:avatar"/><feature var="jabb\
er:iq:browse"/><feature var="jabber:iq:last"/><feature var="jabber:iq:oob"/><f\
eature var="jabber:iq:privacy"/><feature var="jabber:iq:roster"/><feature var=\
"jabber:iq:time"/><feature var="jabber:iq:version"/><feature var="jabber:x:dat\
a"/><feature var="jabber:x:event"/><feature var="jabber:x:oob"/><feature var="\
urn:xmpp:ping"/><feature var="urn:xmpp:time"/><x type="result" xmlns="jabber:x\
:data"><field type="hidden" var="FORM_TYPE"><value>urn:xmpp:dataforms:software\
info</value></field><field var="software"><value>Tkabber</value></field><field\
 var="software_version"><value>1.0-svn-20140122 (Tcl/Tk 8.4.20)</value></field\
><field var="os"><value>FreeBSD</value></field><field var="os_version"><value>\
10.0-STABLE</value></field></x></query>""")
TEST_DB_ENTRY = aioxmpp.xml.read_single_xso(_src, disco.xso.InfoQuery)
TEST_DB_ENTRY_VER = "+0mnUAF1ozCEc37cmdPPsYbsfhg="
TEST_DB_ENTRY_HASH = "sha-1"
TEST_DB_ENTRY_NODE_BARE = "http://tkabber.jabber.ru/"


class TestCache(unittest.TestCase):
    def setUp(self):
        self.c = entitycaps_service.Cache()

    def tearDown(self):
        del self.c

    def test_lookup_in_database_key_errors_if_no_such_entry(self):
        key = unittest.mock.Mock()
        with self.assertRaises(KeyError):
            self.c.lookup_in_database(key)

    def test_system_db_path_used_in_lookup(self):
        base = unittest.mock.Mock()
        base.p = unittest.mock.MagicMock()
        self.c.set_system_db_path(base.p)

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xml.read_single_xso",
                new=base.read_single_xso
            ))

            result = self.c.lookup_in_database(base.key)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.p.__truediv__(base.key.path),
                unittest.mock.call.p.__truediv__().open("rb"),
                unittest.mock.call.p.__truediv__().open().__enter__(),
                unittest.mock.call.read_single_xso(
                    base.p.__truediv__().open(),
                    disco.xso.InfoQuery,
                ),
                unittest.mock.call.p.__truediv__().open().__exit__(
                    None, None, None
                ),
            ]
        )

        self.assertEqual(
            result,
            base.read_single_xso()
        )

    def test_user_db_path_used_in_lookup_as_fallback(self):
        base = unittest.mock.Mock()
        base.p = unittest.mock.MagicMock()
        base.userp = unittest.mock.MagicMock()
        self.c.set_system_db_path(base.p)
        self.c.set_user_db_path(base.userp)

        base.p.__truediv__().open.side_effect = FileNotFoundError()
        base.mock_calls.clear()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xml.read_single_xso",
                new=base.read_single_xso
            ))

            result = self.c.lookup_in_database(base.key)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.p.__truediv__(base.key.path),
                unittest.mock.call.p.__truediv__().open("rb"),
                unittest.mock.call.userp.__truediv__(base.key.path),
                unittest.mock.call.userp.__truediv__().open("rb"),
                unittest.mock.call.userp.__truediv__().open().__enter__(),
                unittest.mock.call.read_single_xso(
                    base.userp.__truediv__().open(),
                    disco.xso.InfoQuery,
                ),
                unittest.mock.call.userp.__truediv__().open().__exit__(
                    None, None, None
                ),
            ]
        )

        self.assertEqual(
            result,
            base.read_single_xso()
        )

    def test_user_db_path_used_if_system_db_is_unset(self):
        base = unittest.mock.Mock()
        base.p = unittest.mock.MagicMock()
        base.userp = unittest.mock.MagicMock()
        self.c.set_user_db_path(base.userp)

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xml.read_single_xso",
                new=base.read_single_xso
            ))

            result = self.c.lookup_in_database(base.key)
        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.userp.__truediv__(base.key.path),
                unittest.mock.call.userp.__truediv__().open("rb"),
                unittest.mock.call.userp.__truediv__().open().__enter__(),
                unittest.mock.call.read_single_xso(
                    base.userp.__truediv__().open(),
                    disco.xso.InfoQuery,
                ),
                unittest.mock.call.userp.__truediv__().open().__exit__(
                    None, None, None
                ),
            ]
        )

        self.assertEqual(
            result,
            base.read_single_xso()
        )

    def test_lookup_uses_lookup_in_database(self):
        with unittest.mock.patch.object(
                self.c,
                "lookup_in_database") as lookup_in_database:
            result = run_coroutine(self.c.lookup(unittest.mock.sentinel.key))

        lookup_in_database.assert_called_with(unittest.mock.sentinel.key)
        self.assertEqual(result, lookup_in_database())

    def test_create_query_future_used_by_lookup(self):
        fut = asyncio.Future()

        base = unittest.mock.Mock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "asyncio.Future",
                new=base.Future
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.c,
                "lookup_in_database",
                new=base.lookup_in_database
            ))

            base.lookup_in_database.side_effect = KeyError()
            base.Future.return_value = fut

            self.assertIs(
                self.c.create_query_future(unittest.mock.sentinel.key),
                fut,
            )

            task = asyncio.ensure_future(
                self.c.lookup(unittest.mock.sentinel.key)
            )
            run_coroutine(asyncio.sleep(0))

            self.assertFalse(task.done())

            fut.set_result(unittest.mock.sentinel.result)

            run_coroutine(asyncio.sleep(0))

            self.assertIs(task.result(), unittest.mock.sentinel.result)

    def test_lookup_key_errors_if_no_matching_entry_or_future(self):
        fut = asyncio.Future()

        base = unittest.mock.Mock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "asyncio.Future",
                new=base.Future
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.c,
                "lookup_in_database",
                new=base.lookup_in_database
            ))

            base.lookup_in_database.side_effect = KeyError()
            base.Future.return_value = fut

            self.assertIs(
                self.c.create_query_future(unittest.mock.sentinel.key),
                fut,
            )

            with self.assertRaises(KeyError):
                run_coroutine(self.c.lookup(unittest.mock.sentinel.other_key))

    def test_lookup_loops_on_query_futures(self):
        fut1 = asyncio.Future()
        fut2 = asyncio.Future()
        fut3 = asyncio.Future()

        base = unittest.mock.Mock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "asyncio.Future",
                new=base.Future
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.c,
                "lookup_in_database",
                new=base.lookup_in_database
            ))

            base.lookup_in_database.side_effect = KeyError()
            base.Future.return_value = fut1
            self.c.create_query_future(unittest.mock.sentinel.key)

            task = asyncio.ensure_future(
                self.c.lookup(unittest.mock.sentinel.key)
            )
            run_coroutine(asyncio.sleep(0))

            self.assertFalse(task.done())

            fut1.set_exception(ValueError())

            base.Future.return_value = fut2
            self.c.create_query_future(unittest.mock.sentinel.key)

            run_coroutine(asyncio.sleep(0))

            self.assertFalse(task.done())

            fut2.set_exception(ValueError())

            base.Future.return_value = fut3
            self.c.create_query_future(unittest.mock.sentinel.key)

            run_coroutine(asyncio.sleep(0))

            self.assertFalse(task.done())

            fut3.set_result(base.result)

            run_coroutine(asyncio.sleep(0))

            self.assertIs(task.result(), base.result)

    def test_lookup_key_errors_if_last_query_future_fails(self):
        fut = asyncio.Future()

        base = unittest.mock.Mock()
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "asyncio.Future",
                new=base.Future
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.c,
                "lookup_in_database",
                new=base.lookup_in_database
            ))

            base.lookup_in_database.side_effect = KeyError()
            base.Future.return_value = fut

            self.assertIs(
                self.c.create_query_future(unittest.mock.sentinel.key),
                fut,
            )

            task = asyncio.ensure_future(
                self.c.lookup(unittest.mock.sentinel.key)
            )
            run_coroutine(asyncio.sleep(0))

            self.assertFalse(task.done())

            fut.set_exception(ValueError())

            run_coroutine(asyncio.sleep(0))

            with self.assertRaises(KeyError):
                run_coroutine(task)

    def test_add_cache_entry_is_immediately_visible_in_lookup_and_defers_writeback(self):  # NOQA
        q = disco.xso.InfoQuery()
        p = unittest.mock.MagicMock()
        key = unittest.mock.Mock()
        self.c.set_user_db_path(p)

        with contextlib.ExitStack() as stack:
            copy = stack.enter_context(unittest.mock.patch(
                "copy.copy"
            ))

            run_in_executor = stack.enter_context(unittest.mock.patch.object(
                asyncio.get_event_loop(),
                "run_in_executor"
            ))

            async_ = stack.enter_context(unittest.mock.patch(
                "asyncio.ensure_future"
            ))

            self.c.add_cache_entry(
                key,
                q,
            )

        copy.assert_called_once_with(q)

        p.__truediv__.assert_called_once_with(key.path)

        run_in_executor.assert_called_with(
            None,
            entitycaps_service.writeback,
            p.__truediv__(),
            q.captured_events,
        )
        async_.assert_called_with(run_in_executor())

        result = self.c.lookup_in_database(key)
        self.assertEqual(result, copy())

    def test_add_cache_entry_does_not_perform_writeback_if_no_userdb_is_set(self):  # NOQA
        q = disco.xso.InfoQuery()

        with contextlib.ExitStack() as stack:
            copy = stack.enter_context(unittest.mock.patch(
                "copy.copy"
            ))

            run_in_executor = stack.enter_context(unittest.mock.patch.object(
                asyncio.get_event_loop(),
                "run_in_executor"
            ))

            async_ = stack.enter_context(unittest.mock.patch(
                "asyncio.ensure_future"
            ))

            self.c.add_cache_entry(
                unittest.mock.sentinel.key,
                q,
            )

        copy.assert_called_with(q)
        self.assertFalse(run_in_executor.mock_calls)
        self.assertFalse(async_.mock_calls)

        result = self.c.lookup_in_database(unittest.mock.sentinel.key)
        self.assertEqual(result, copy())

    def test_writeback_from_add_cache_entry_creates_file(self):
        q = disco.xso.InfoQuery()
        q.captured_events = [
            ("start", q.TAG[0], q.TAG[1], {}),
            ("end",)
        ]
        key = unittest.mock.Mock()

        with contextlib.ExitStack() as stack:
            tempdir = stack.enter_context(
                tempfile.TemporaryDirectory()
            )

            p = pathlib.Path(tempdir)
            key.path = pathlib.Path("key") / "path" / "used"

            self.c.set_user_db_path(p)

            copy = stack.enter_context(unittest.mock.patch(
                "copy.copy"
            ))

            run_in_executor = stack.enter_context(unittest.mock.patch.object(
                asyncio.get_event_loop(),
                "run_in_executor"
            ))

            async_ = stack.enter_context(unittest.mock.patch(
                "asyncio.ensure_future"
            ))

            self.c.add_cache_entry(
                key,
                q,
            )

            copy.assert_called_once_with(q)

            run_in_executor.assert_called_with(
                None,
                entitycaps_service.writeback,
                p / key.path,
                q.captured_events,
            )
            async_.assert_called_with(run_in_executor())

            entitycaps_service.writeback(
                *run_in_executor.mock_calls[0][1][2:]
            )

            self.assertTrue((p / key.path).is_file())


class TestService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.disco_client = unittest.mock.Mock()
        self.disco_client.query_info = CoroutineMock()
        self.disco_client.query_info.side_effect = AssertionError()
        self.disco_server = unittest.mock.Mock()
        self.disco_server.on_info_changed.context_connect = \
            unittest.mock.MagicMock()
        self.disco_server.iter_items.return_value = []

        self.impl115 = unittest.mock.Mock()
        self.impl115.extract_keys.return_value = []
        self.impl115.calculate_keys.return_value = []

        self.impl390 = unittest.mock.Mock()
        self.impl390.extract_keys.return_value = []
        self.impl390.calculate_keys.return_value = []

        with contextlib.ExitStack() as stack:
            Implementation115 = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps115.Implementation"
                )
            )
            Implementation115.return_value = self.impl115

            Implementation390 = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps390.Implementation"
                )
            )
            Implementation390.return_value = self.impl390

            self.s = entitycaps_service.EntityCapsService(
                self.cc,
                dependencies={
                    disco.DiscoClient: self.disco_client,
                    disco.DiscoServer: self.disco_server,
                }
            )

        Implementation115.assert_called_once_with(
            entitycaps_service.EntityCapsService.NODE
        )

        Implementation390.assert_called_once_with(
            aioxmpp.hashes.default_hash_algorithms
        )

        self.disco_client.mock_calls.clear()
        self.disco_server.mock_calls.clear()
        self.cc.mock_calls.clear()

        self.disco_server.iter_features.return_value = [
            "http://jabber.org/protocol/disco#items",
            "http://jabber.org/protocol/disco#info",
        ]

        self.disco_server.iter_identities.return_value = [
            ("client", "pc", None, None),
            ("client", "pc", structs.LanguageTag.fromstr("en"), "foo"),
        ]

    def test_is_Service_subclass(self):
        self.assertTrue(issubclass(
            entitycaps_service.EntityCapsService,
            service.Service
        ))

    def test_registers_xep115_feature(self):
        self.assertIsInstance(
            entitycaps_service.EntityCapsService._xep115_feature,
            aioxmpp.disco.register_feature,
        )
        self.assertEqual(
            entitycaps_service.EntityCapsService._xep115_feature.feature,
            namespaces.xep0115_caps,
        )

    def test_registers_xep390_feature(self):
        self.assertIsInstance(
            entitycaps_service.EntityCapsService._xep390_feature,
            aioxmpp.disco.register_feature,
        )
        self.assertEqual(
            entitycaps_service.EntityCapsService._xep390_feature.feature,
            namespaces.xep0390_caps,
        )

    def test_xep115_support_sets_xep115_feature_enabledness(self):
        self.assertTrue(self.s.xep115_support)
        self.assertTrue(self.s._xep115_feature.enabled)

        self.s.xep115_support = False

        self.assertFalse(self.s.xep115_support)
        self.assertFalse(self.s._xep115_feature.enabled)

        self.s.xep115_support = True

        self.assertTrue(self.s.xep115_support)
        self.assertTrue(self.s._xep115_feature.enabled)

    def test_xep390_support_sets_xep390_feature_enabledness(self):
        self.assertTrue(self.s.xep390_support)
        self.assertTrue(self.s._xep390_feature.enabled)

        self.s.xep390_support = False

        self.assertFalse(self.s.xep390_support)
        self.assertFalse(self.s._xep390_feature.enabled)

        self.s.xep390_support = True

        self.assertTrue(self.s.xep390_support)
        self.assertTrue(self.s._xep390_feature.enabled)

    def test_after_disco(self):
        self.assertLess(
            disco.DiscoServer,
            entitycaps_service.EntityCapsService
        )
        self.assertLess(
            disco.DiscoClient,
            entitycaps_service.EntityCapsService
        )

    def test_handle_outbound_presence_is_decorated(self):
        self.assertTrue(
            service.is_outbound_presence_filter(
                entitycaps_service.EntityCapsService.handle_outbound_presence,
            )
        )

    def test_handle_inbound_presence_is_decorated(self):
        self.assertTrue(
            service.is_inbound_presence_filter(
                entitycaps_service.EntityCapsService.handle_inbound_presence,
            )
        )

    def test_cache_defaults(self):
        self.assertIsInstance(
            self.s.cache,
            entitycaps_service.Cache
        )

    def test_cache_defaults_when_deleted(self):
        c1 = self.s.cache
        del self.s.cache
        c2 = self.s.cache
        self.assertIsNot(c1, c2)
        self.assertIsInstance(c2, entitycaps_service.Cache)

    def test_cache_can_be_set(self):
        c = unittest.mock.Mock()
        self.s.cache = c
        self.assertIs(self.s.cache, c)

    def test_handle_inbound_presence_extracts_115_keys_and_spawns_lookup(self):
        presence = unittest.mock.Mock(spec=aioxmpp.Presence)

        self.impl115.extract_keys.return_value = iter([
            unittest.mock.sentinel.key1,
        ])

        with contextlib.ExitStack() as stack:
            LazyTask = stack.enter_context(
                unittest.mock.patch("aioxmpp.utils.LazyTask")
            )

            lookup_info = stack.enter_context(
                unittest.mock.patch.object(self.s, "lookup_info")
            )

            result = self.s.handle_inbound_presence(presence)

        self.impl115.extract_keys.assert_called_once_with(presence)

        LazyTask.assert_called_once_with(
            lookup_info,
            presence.from_,
            [
                unittest.mock.sentinel.key1,
            ]
        )

        self.disco_client.set_info_future.assert_called_with(
            presence.from_,
            None,
            LazyTask(),
        )

        self.assertEqual(result, presence)

    def test_handle_inbound_presence_extracts_390_keys_and_spawns_lookup(self):
        presence = unittest.mock.Mock(spec=aioxmpp.Presence)

        self.impl390.extract_keys.return_value = iter([
            unittest.mock.sentinel.key1,
        ])

        with contextlib.ExitStack() as stack:
            LazyTask = stack.enter_context(
                unittest.mock.patch("aioxmpp.utils.LazyTask")
            )

            lookup_info = stack.enter_context(
                unittest.mock.patch.object(self.s, "lookup_info")
            )

            result = self.s.handle_inbound_presence(presence)

        self.impl390.extract_keys.assert_called_once_with(presence)

        LazyTask.assert_called_once_with(
            lookup_info,
            presence.from_,
            [
                unittest.mock.sentinel.key1,
            ]
        )

        self.disco_client.set_info_future.assert_called_with(
            presence.from_,
            None,
            LazyTask(),
        )

        self.assertEqual(result, presence)

    def test_handle_inbound_presence_extracts_both_keys_and_spawns_lookup(self):  # NOQA
        presence = unittest.mock.Mock(spec=aioxmpp.Presence)

        self.impl115.extract_keys.return_value = iter([
            unittest.mock.sentinel.key1,
        ])

        self.impl390.extract_keys.return_value = iter([
            unittest.mock.sentinel.key2,
            unittest.mock.sentinel.key3,
        ])

        with contextlib.ExitStack() as stack:
            LazyTask = stack.enter_context(
                unittest.mock.patch("aioxmpp.utils.LazyTask")
            )

            lookup_info = stack.enter_context(
                unittest.mock.patch.object(self.s, "lookup_info")
            )

            result = self.s.handle_inbound_presence(presence)

        self.impl390.extract_keys.assert_called_once_with(presence)

        LazyTask.assert_called_once_with(
            lookup_info,
            presence.from_,
            [
                unittest.mock.sentinel.key2,
                unittest.mock.sentinel.key3,
                unittest.mock.sentinel.key1,
            ]
        )

        self.disco_client.set_info_future.assert_called_with(
            presence.from_,
            None,
            LazyTask(),
        )

        self.assertEqual(result, presence)

    def test_handle_inbound_presence_ignores_115_if_disabled(self):
        self.s.xep115_support = False

        presence = unittest.mock.Mock(spec=aioxmpp.Presence)

        self.impl115.extract_keys.return_value = iter([
            unittest.mock.sentinel.key1,
        ])

        self.impl390.extract_keys.return_value = iter([
            unittest.mock.sentinel.key2,
            unittest.mock.sentinel.key3,
        ])

        with contextlib.ExitStack() as stack:
            LazyTask = stack.enter_context(
                unittest.mock.patch("aioxmpp.utils.LazyTask")
            )

            lookup_info = stack.enter_context(
                unittest.mock.patch.object(self.s, "lookup_info")
            )

            result = self.s.handle_inbound_presence(presence)

        self.impl115.extract_keys.assert_not_called()
        self.impl390.extract_keys.assert_called_once_with(presence)

        LazyTask.assert_called_once_with(
            lookup_info,
            presence.from_,
            [
                unittest.mock.sentinel.key2,
                unittest.mock.sentinel.key3,
            ]
        )

        self.disco_client.set_info_future.assert_called_with(
            presence.from_,
            None,
            LazyTask(),
        )

        self.assertEqual(result, presence)

    def test_handle_inbound_presence_ignores_390_if_disabled(self):
        self.s.xep390_support = False

        presence = unittest.mock.Mock(spec=aioxmpp.Presence)

        self.impl115.extract_keys.return_value = iter([
            unittest.mock.sentinel.key1,
        ])

        self.impl390.extract_keys.return_value = iter([
            unittest.mock.sentinel.key2,
            unittest.mock.sentinel.key3,
        ])

        with contextlib.ExitStack() as stack:
            LazyTask = stack.enter_context(
                unittest.mock.patch("aioxmpp.utils.LazyTask")
            )

            lookup_info = stack.enter_context(
                unittest.mock.patch.object(self.s, "lookup_info")
            )

            result = self.s.handle_inbound_presence(presence)

        self.impl115.extract_keys.assert_called_once_with(presence)
        self.impl390.extract_keys.assert_not_called()

        LazyTask.assert_called_once_with(
            lookup_info,
            presence.from_,
            [
                unittest.mock.sentinel.key1,
            ]
        )

        self.disco_client.set_info_future.assert_called_with(
            presence.from_,
            None,
            LazyTask(),
        )

        self.assertEqual(result, presence)

    def test_query_and_cache(self):
        self.maxDiff = None

        ver = TEST_DB_ENTRY_VER
        response = TEST_DB_ENTRY

        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.key.verify.return_value = True
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "add_cache_entry",
                new=base.add_cache_entry
            ))

            base.disco.query_info.return_value = response

            result = run_coroutine(self.s.query_and_cache(
                TEST_FROM,
                base.key,
                base.fut,
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.disco.query_info(
                    TEST_FROM,
                    node=base.key.node,
                    require_fresh=True,
                    no_cache=True,
                ),
                unittest.mock.call.key.verify(response),
                unittest.mock.call.add_cache_entry(
                    base.key,
                    response,
                ),
                unittest.mock.call.fut.set_result(
                    result,
                )
            ]
        )

        self.assertEqual(result, response)

    def test_query_and_cache_checks_hash(self):
        self.maxDiff = None

        ver = TEST_DB_ENTRY_VER
        response = TEST_DB_ENTRY

        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.key.verify.return_value = False
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "add_cache_entry",
                new=base.add_cache_entry
            ))

            base.disco.query_info.return_value = response

            result = run_coroutine(self.s.query_and_cache(
                TEST_FROM,
                base.key,
                base.fut,
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.disco.query_info(
                    TEST_FROM,
                    node=base.key.node,
                    require_fresh=True,
                    no_cache=True,
                ),
                unittest.mock.call.key.verify(response),
                unittest.mock.call.fut.set_exception(unittest.mock.ANY)
            ]
        )

        _, (exc,), _ = base.fut.mock_calls[0]
        self.assertIsInstance(exc, ValueError)

        self.assertEqual(result, response)

    def test_query_and_cache_does_not_cache_on_ValueError(self):
        self.maxDiff = None

        response = TEST_DB_ENTRY
        exc = ValueError()

        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.key.verify.side_effect = exc
        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "add_cache_entry",
                new=base.add_cache_entry
            ))

            base.disco.query_info.return_value = response

            result = run_coroutine(self.s.query_and_cache(
                TEST_FROM,
                base.key,
                base.fut,
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.disco.query_info(
                    TEST_FROM,
                    node=base.key.node,
                    require_fresh=True,
                    no_cache=True,
                ),
                unittest.mock.call.key.verify(response),
                unittest.mock.call.fut.set_exception(exc)
            ]
        )

        self.assertEqual(result, response)

    def test_lookup_info_asks_cache_first_and_returns_value(self):
        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.query_and_cache = CoroutineMock()
        base.lookup = CoroutineMock()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "lookup",
                new=base.lookup,
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.s,
                "query_and_cache",
                new=base.query_and_cache
            ))

            base.lookup.return_value = unittest.mock.sentinel.cache_result

            result = run_coroutine(self.s.lookup_info(
                TEST_FROM,
                [
                    unittest.mock.sentinel.key,
                ]
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.lookup(unittest.mock.sentinel.key),
            ]
        )

        self.assertIs(result, unittest.mock.sentinel.cache_result)

    def test_lookup_info_asks_cache_for_all_keys_and_returns_value(self):
        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.query_and_cache = CoroutineMock()
        base.lookup = CoroutineMock()

        ncall = 0

        def lookup_side_effect(*args, **kwargs):
            nonlocal ncall
            ncall += 1
            if ncall == 2:
                return unittest.mock.sentinel.cache_result
            raise KeyError()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "lookup",
                new=base.lookup,
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.s,
                "query_and_cache",
                new=base.query_and_cache
            ))

            base.lookup.side_effect = lookup_side_effect

            result = run_coroutine(self.s.lookup_info(
                TEST_FROM,
                [
                    unittest.mock.sentinel.key1,
                    unittest.mock.sentinel.key2,
                ]
            ))

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.lookup(unittest.mock.sentinel.key1),
                unittest.mock.call.lookup(unittest.mock.sentinel.key2),
            ]
        )

        self.assertIs(result, unittest.mock.sentinel.cache_result)

    def test_lookup_info_delegates_to_query_and_cache_on_miss(self):
        base = unittest.mock.Mock()
        base.disco = self.disco_client
        base.disco.query_info.side_effect = None
        base.query_and_cache = CoroutineMock()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "lookup",
                new=base.lookup
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.s.cache,
                "create_query_future",
                new=base.create_query_future
            ))

            stack.enter_context(unittest.mock.patch.object(
                self.s,
                "query_and_cache",
                new=base.query_and_cache
            ))

            base.lookup.side_effect = KeyError()
            base.query_and_cache.return_value = \
                unittest.mock.sentinel.query_result

            result = run_coroutine(self.s.lookup_info(
                TEST_FROM,
                [
                    unittest.mock.sentinel.key1,
                    unittest.mock.sentinel.key2,
                ]
            ))

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.lookup(unittest.mock.sentinel.key1),
                unittest.mock.call.lookup(unittest.mock.sentinel.key2),
                unittest.mock.call.create_query_future(
                    unittest.mock.sentinel.key1,
                ),
                unittest.mock.call.query_and_cache(
                    TEST_FROM,
                    unittest.mock.sentinel.key1,
                    base.create_query_future()
                )
            ]
        )

        self.assertIs(result, unittest.mock.sentinel.query_result)

    def test_update_hash(self):
        base = unittest.mock.Mock()

        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])

        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        with contextlib.ExitStack() as stack:
            hash_query = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps115.hash_query"
                )
            )

            clone = stack.enter_context(
                unittest.mock.patch.object(
                    disco.StaticNode,
                    "clone",
                )
            )

            push_hashset = stack.enter_context(
                unittest.mock.patch.object(
                    self.s,
                    "_push_hashset",
                )
            )
            push_hashset.return_value = True

            with self.s.on_ver_changed.context_connect(base.cb):
                self.s.update_hash()

        hash_query.assert_not_called()

        clone.assert_called_once_with(self.disco_server)
        clone().as_info_xso.assert_called_once_with()

        self.impl115.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.impl390.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.disco_server.mount_node.assert_not_called()

        push_hashset.assert_called_once_with(
            clone(),
            {
                self.impl115: {base.key1},
                self.impl390: {base.key2, base.key3},
            }
        )

        base.cb.assert_called_once_with()

    def test_update_hash_does_not_emit_on_ver_changed_if_push_hashset_returns_false(self):  # NOQA
        base = unittest.mock.Mock()

        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])

        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        with contextlib.ExitStack() as stack:
            hash_query = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps115.hash_query"
                )
            )

            clone = stack.enter_context(
                unittest.mock.patch.object(
                    disco.StaticNode,
                    "clone",
                )
            )

            push_hashset = stack.enter_context(
                unittest.mock.patch.object(
                    self.s,
                    "_push_hashset",
                )
            )
            push_hashset.return_value = False

            with self.s.on_ver_changed.context_connect(base.cb):
                self.s.update_hash()

        hash_query.assert_not_called()

        clone.assert_called_once_with(self.disco_server)
        clone().as_info_xso.assert_called_once_with()

        self.impl115.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.impl390.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.disco_server.mount_node.assert_not_called()

        push_hashset.assert_called_once_with(
            clone(),
            {
                self.impl115: {base.key1},
                self.impl390: {base.key2, base.key3},
            }
        )

        base.cb.assert_not_called()

    def test_update_hash_ignores_115_if_disabled(self):
        self.s.xep115_support = False
        self.disco_server.reset_mock()

        base = unittest.mock.Mock()

        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])

        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        with contextlib.ExitStack() as stack:
            hash_query = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps115.hash_query"
                )
            )

            clone = stack.enter_context(
                unittest.mock.patch.object(
                    disco.StaticNode,
                    "clone",
                )
            )

            push_hashset = stack.enter_context(
                unittest.mock.patch.object(
                    self.s,
                    "_push_hashset",
                )
            )

            self.s.update_hash()

        hash_query.assert_not_called()

        clone.assert_called_once_with(self.disco_server)
        clone().as_info_xso.assert_called_once_with()

        self.impl115.calculate_keys.assert_not_called()

        self.impl390.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.disco_server.mount_node.assert_not_called()

        push_hashset.assert_called_once_with(
            clone(),
            {
                self.impl390: {base.key2, base.key3},
            }
        )

    def test_update_hash_ignores_390_if_disabled(self):
        self.s.xep390_support = False
        self.disco_server.reset_mock()

        base = unittest.mock.Mock()

        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])

        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        with contextlib.ExitStack() as stack:
            hash_query = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.entitycaps.caps115.hash_query"
                )
            )

            clone = stack.enter_context(
                unittest.mock.patch.object(
                    disco.StaticNode,
                    "clone",
                )
            )

            push_hashset = stack.enter_context(
                unittest.mock.patch.object(
                    self.s,
                    "_push_hashset",
                )
            )

            self.s.update_hash()

        hash_query.assert_not_called()

        clone.assert_called_once_with(self.disco_server)
        clone().as_info_xso.assert_called_once_with()

        self.impl115.calculate_keys.assert_called_once_with(
            clone().as_info_xso()
        )

        self.impl390.calculate_keys.assert_not_called()

        self.disco_server.mount_node.assert_not_called()

        push_hashset.assert_called_once_with(
            clone(),
            {
                self.impl115: {base.key1},
            }
        )

    def test__push_hashset_registers_nodes(self):
        base = unittest.mock.Mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key2,
                    base.key3,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node,
                                   unittest.mock.sentinel.node),
                unittest.mock.call(base.key2.node,
                                   unittest.mock.sentinel.node),
                unittest.mock.call(base.key3.node,
                                   unittest.mock.sentinel.node),
            ]
        )

    def test__push_hashset_keeps_three_hashsets(self):
        base = unittest.mock.Mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node1,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key2,
                    base.key3,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node2,
            {
                unittest.mock.sentinel.impl1: {
                    base.key4,
                },
                unittest.mock.sentinel.impl2: {
                    base.key5,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node3,
            {
                unittest.mock.sentinel.impl1: {
                    base.key6,
                },
                unittest.mock.sentinel.impl2: {
                    base.key7,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key2.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key3.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key4.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key5.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key6.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key7.node,
                                   unittest.mock.sentinel.node3),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [],
        )

    def test__push_hashset_fifo_behaviour(self):
        base = unittest.mock.Mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node1,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key2,
                    base.key3,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node2,
            {
                unittest.mock.sentinel.impl1: {
                    base.key4,
                },
                unittest.mock.sentinel.impl2: {
                    base.key5,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node3,
            {
                unittest.mock.sentinel.impl1: {
                    base.key6,
                },
                unittest.mock.sentinel.impl2: {
                    base.key7,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node4,
            {
                unittest.mock.sentinel.impl1: {
                    base.key8,
                },
                unittest.mock.sentinel.impl2: {
                    base.key9,
                    base.key10,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key2.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key3.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key4.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key5.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key6.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key7.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key8.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key9.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key10.node,
                                   unittest.mock.sentinel.node4),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node),
                unittest.mock.call(base.key2.node),
                unittest.mock.call(base.key3.node),
            ],
        )

        self.disco_server.unmount_node.reset_mock()
        self.disco_server.mount_node.reset_mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node5,
            {
                unittest.mock.sentinel.impl1: {
                    base.key11,
                },
                unittest.mock.sentinel.impl2: {
                    base.key12,
                    base.key13,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key11.node,
                                   unittest.mock.sentinel.node5),
                unittest.mock.call(base.key12.node,
                                   unittest.mock.sentinel.node5),
                unittest.mock.call(base.key13.node,
                                   unittest.mock.sentinel.node5),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [
                unittest.mock.call(base.key4.node),
                unittest.mock.call(base.key5.node),
            ],
        )

    def test__push_hashset_can_deal_with_hash_collisions(self):
        # I know we’re insane to even consider this…
        base = unittest.mock.Mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node1,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key2,
                    base.key3,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node2,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key5,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node3,
            {
                unittest.mock.sentinel.impl1: {
                    base.key6,
                },
                unittest.mock.sentinel.impl2: {
                    base.key7,
                }
            }
        )

        self.s._push_hashset(
            unittest.mock.sentinel.node4,
            {
                unittest.mock.sentinel.impl1: {
                    base.key8,
                },
                unittest.mock.sentinel.impl2: {
                    base.key9,
                    base.key10,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key2.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key3.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key5.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key6.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key7.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key8.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key9.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key10.node,
                                   unittest.mock.sentinel.node4),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [
                unittest.mock.call(base.key2.node),
                unittest.mock.call(base.key3.node),
            ],
        )

        self.disco_server.unmount_node.reset_mock()
        self.disco_server.mount_node.reset_mock()

        self.s._push_hashset(
            unittest.mock.sentinel.node5,
            {
                unittest.mock.sentinel.impl1: {
                    base.key11,
                },
                unittest.mock.sentinel.impl2: {
                    base.key12,
                    base.key13,
                }
            }
        )

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key11.node,
                                   unittest.mock.sentinel.node5),
                unittest.mock.call(base.key12.node,
                                   unittest.mock.sentinel.node5),
                unittest.mock.call(base.key13.node,
                                   unittest.mock.sentinel.node5),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node),
                unittest.mock.call(base.key5.node),
            ],
        )

    def test__push_hashset_noop_and_returs_false_on_dup(self):
        base = unittest.mock.Mock()

        self.assertTrue(self.s._push_hashset(
            unittest.mock.sentinel.node1,
            {
                unittest.mock.sentinel.impl1: {
                    base.key1,
                },
                unittest.mock.sentinel.impl2: {
                    base.key2,
                    base.key3,
                }
            }
        ))

        self.assertTrue(self.s._push_hashset(
            unittest.mock.sentinel.node2,
            {
                unittest.mock.sentinel.impl1: {
                    base.key4,
                },
                unittest.mock.sentinel.impl2: {
                    base.key5,
                }
            }
        ))

        self.assertTrue(self.s._push_hashset(
            unittest.mock.sentinel.node3,
            {
                unittest.mock.sentinel.impl1: {
                    base.key6,
                },
                unittest.mock.sentinel.impl2: {
                    base.key7,
                }
            }
        ))

        self.assertTrue(self.s._push_hashset(
            unittest.mock.sentinel.node4,
            {
                unittest.mock.sentinel.impl1: {
                    base.key8,
                },
                unittest.mock.sentinel.impl2: {
                    base.key9,
                    base.key10,
                }
            }
        ))

        self.assertFalse(self.s._push_hashset(
            unittest.mock.sentinel.node5,
            {
                unittest.mock.sentinel.impl1: {
                    base.key8,
                },
                unittest.mock.sentinel.impl2: {
                    base.key9,
                    base.key10,
                }
            }
        ))

        self.assertCountEqual(
            self.disco_server.mount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key2.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key3.node,
                                   unittest.mock.sentinel.node1),
                unittest.mock.call(base.key4.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key5.node,
                                   unittest.mock.sentinel.node2),
                unittest.mock.call(base.key6.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key7.node,
                                   unittest.mock.sentinel.node3),
                unittest.mock.call(base.key8.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key9.node,
                                   unittest.mock.sentinel.node4),
                unittest.mock.call(base.key10.node,
                                   unittest.mock.sentinel.node4),
            ]
        )

        self.assertCountEqual(
            self.disco_server.unmount_node.mock_calls,
            [
                unittest.mock.call(base.key1.node),
                unittest.mock.call(base.key2.node),
                unittest.mock.call(base.key3.node),
            ],
        )

    def test__info_changed_calls_update_hash_soon(self):
        with contextlib.ExitStack() as stack:
            get_event_loop = stack.enter_context(unittest.mock.patch(
                "asyncio.get_event_loop"
            ))

            self.s._info_changed()

        get_event_loop.assert_called_with()
        get_event_loop().call_soon.assert_called_with(
            self.s.update_hash
        )

    def test_handle_outbound_presence_inserts_keys(self):
        base = unittest.mock.Mock()
        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])
        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        self.s._push_hashset(
            unittest.mock.sentinel.node,
            {
                self.impl115: {base.key1},
                self.impl390: {base.key2, base.key3},
            }
        )

        presence = stanza.Presence()
        result = self.s.handle_outbound_presence(presence)
        self.assertIs(result, presence)

        self.impl115.put_keys.assert_called_once_with(
            {base.key1},
            presence,
        )

        self.impl390.put_keys.assert_called_once_with(
            {base.key2, base.key3},
            presence,
        )

    def test_handle_outbound_presence_does_not_attach_caps_to_non_available(
            self):
        base = unittest.mock.Mock()
        self.impl115.calculate_keys.return_value = iter([
            base.key1,
        ])
        self.impl390.calculate_keys.return_value = iter([
            base.key2,
            base.key3,
        ])

        self.s._push_hashset(
            unittest.mock.sentinel.node,
            {
                self.impl115: {base.key1},
                self.impl390: {base.key2, base.key3},
            }
        )

        types = [
            structs.PresenceType.UNAVAILABLE,
            structs.PresenceType.SUBSCRIBE,
            structs.PresenceType.SUBSCRIBED,
            structs.PresenceType.UNSUBSCRIBE,
            structs.PresenceType.UNSUBSCRIBED,
            structs.PresenceType.ERROR,
        ]

        for type_ in types:
            presence = stanza.Presence(type_=type_)
            self.s.handle_outbound_presence(presence)

        self.impl115.put_keys.assert_not_called()

    def test_handle_outbound_presence_does_not_call_put_keys_if_unset(self):
        presence = stanza.Presence()
        self.s.handle_outbound_presence(presence)
        self.impl115.put_keys.assert_not_called()


class Testwriteback(unittest.TestCase):
    def test_uses_tempfile_atomically_and_serialises_xso(self):
        base = unittest.mock.Mock()
        base.p = unittest.mock.MagicMock()
        base.NamedTemporaryFile = unittest.mock.MagicMock()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "tempfile.NamedTemporaryFile",
                new=base.NamedTemporaryFile
            ))

            stack.enter_context(unittest.mock.patch(
                "os.replace",
                new=base.replace
            ))

            stack.enter_context(unittest.mock.patch(
                "os.unlink",
                new=base.unlink
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xml.XMPPXMLGenerator",
                new=base.XMPPXMLGenerator
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xso.events_to_sax",
                new=base.events_to_sax
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.utils.mkdir_exist_ok",
                new=base.mkdir_exist_ok,
            ))

            entitycaps_service.writeback(base.p,
                                         base.data)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.mkdir_exist_ok(base.p.parent),
                unittest.mock._Call(("p.parent.__str__", (), {})),
                unittest.mock.call.NamedTemporaryFile(
                    dir=str(base.p.parent),
                    delete=False
                ),
                unittest.mock.call.NamedTemporaryFile().__enter__(),
                unittest.mock.call.XMPPXMLGenerator(
                    base.NamedTemporaryFile().__enter__(),
                    short_empty_elements=True,
                ),
                unittest.mock.call.XMPPXMLGenerator().startDocument(),
                unittest.mock.call.events_to_sax(
                    base.data,
                    base.XMPPXMLGenerator()
                ),
                unittest.mock.call.XMPPXMLGenerator().endDocument(),
                unittest.mock.call.NamedTemporaryFile().__exit__(
                    None, None, None
                ),
                unittest.mock._Call(("p.__str__", (), {})),
                unittest.mock.call.replace(
                    base.NamedTemporaryFile().__enter__().name,
                    str(base.p),
                ),
            ]
        )

    def test_unlinks_tempfile_on_error(self):
        base = unittest.mock.Mock()
        base.p = unittest.mock.MagicMock()
        base.NamedTemporaryFile = unittest.mock.MagicMock()

        with contextlib.ExitStack() as stack:
            stack.enter_context(unittest.mock.patch(
                "tempfile.NamedTemporaryFile",
                new=base.NamedTemporaryFile
            ))

            stack.enter_context(unittest.mock.patch(
                "os.replace",
                new=base.replace
            ))

            stack.enter_context(unittest.mock.patch(
                "os.unlink",
                new=base.unlink
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xml.XMPPXMLGenerator",
                new=base.XMPPXMLGenerator
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.xso.events_to_sax",
                new=base.events_to_sax
            ))

            stack.enter_context(unittest.mock.patch(
                "aioxmpp.utils.mkdir_exist_ok",
                new=base.mkdir_exist_ok,
            ))

            exc = Exception()
            base.events_to_sax.side_effect = exc

            with self.assertRaises(Exception) as ctx:
                entitycaps_service.writeback(
                    base.p,
                    base.data)

        self.assertIs(ctx.exception, exc)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.mkdir_exist_ok(base.p.parent),
                unittest.mock._Call(("p.parent.__str__", (), {})),
                unittest.mock.call.NamedTemporaryFile(
                    dir=str(base.p.parent),
                    delete=False
                ),
                unittest.mock.call.NamedTemporaryFile().__enter__(),
                unittest.mock.call.XMPPXMLGenerator(
                    base.NamedTemporaryFile().__enter__(),
                    short_empty_elements=True,
                ),
                unittest.mock.call.XMPPXMLGenerator().startDocument(),
                unittest.mock.call.events_to_sax(
                    base.data,
                    base.XMPPXMLGenerator()
                ),
                unittest.mock.call.unlink(
                    base.NamedTemporaryFile().__enter__().name
                ),
                unittest.mock.call.NamedTemporaryFile().__exit__(
                    unittest.mock.ANY,
                    unittest.mock.ANY,
                    unittest.mock.ANY,
                ),
            ]
        )

# foo



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.xso as xso
import aioxmpp.ibr.xso as ibr_xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0077_in_band,
            "jabber:iq:register"
        )


class TestQuery(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            ibr_xso.Query,
            aioxmpp.xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            ibr_xso.Query.TAG,
            (namespaces.xep0077_in_band, "query")
        )

    def test_is_iq_payload(self):
        self.assertIn(
            ibr_xso.Query.TAG,
            aioxmpp.IQ.CHILD_MAP,
        )

    def test_username(self):
        self.assertIsInstance(
            ibr_xso.Query.username.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.username.tag,
            (namespaces.xep0077_in_band, "username")
        )
        self.assertIsNone(
            ibr_xso.Query.username.default,
        )

    def test_nick(self):
        self.assertIsInstance(
            ibr_xso.Query.nick.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.nick.tag,
            (namespaces.xep0077_in_band, "nick")
        )
        self.assertIsNone(
            ibr_xso.Query.nick.default,
        )

    def test_password(self):
        self.assertIsInstance(
            ibr_xso.Query.password.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.password.tag,
            (namespaces.xep0077_in_band, "password")
        )
        self.assertIsNone(
            ibr_xso.Query.password.default,
        )

    def test_name(self):
        self.assertIsInstance(
            ibr_xso.Query.name.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.name.tag,
            (namespaces.xep0077_in_band, "name")
        )
        self.assertIsNone(
            ibr_xso.Query.name.default,
        )

    def test_first(self):
        self.assertIsInstance(
            ibr_xso.Query.first.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.first.tag,
            (namespaces.xep0077_in_band, "first")
        )
        self.assertIsNone(
            ibr_xso.Query.first.default,
        )

    def test_last(self):
        self.assertIsInstance(
            ibr_xso.Query.last.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.last.tag,
            (namespaces.xep0077_in_band, "last")
        )
        self.assertIsNone(
            ibr_xso.Query.last.default,
        )

    def test_email(self):
        self.assertIsInstance(
            ibr_xso.Query.email.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.email.tag,
            (namespaces.xep0077_in_band, "email")
        )
        self.assertIsNone(
            ibr_xso.Query.email.default,
        )

    def test_address(self):
        self.assertIsInstance(
            ibr_xso.Query.address.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.address.tag,
            (namespaces.xep0077_in_band, "address")
        )
        self.assertIsNone(
            ibr_xso.Query.address.default,
        )

    def test_city(self):
        self.assertIsInstance(
            ibr_xso.Query.city.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.city.tag,
            (namespaces.xep0077_in_band, "city")
        )
        self.assertIsNone(
            ibr_xso.Query.city.default,
        )

    def test_state(self):
        self.assertIsInstance(
            ibr_xso.Query.state.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.state.tag,
            (namespaces.xep0077_in_band, "state")
        )
        self.assertIsNone(
            ibr_xso.Query.state.default,
        )

    def test_zip(self):
        self.assertIsInstance(
            ibr_xso.Query.zip.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.zip.tag,
            (namespaces.xep0077_in_band, "zip")
        )
        self.assertIsNone(
            ibr_xso.Query.zip.default,
        )

    def test_phone(self):
        self.assertIsInstance(
            ibr_xso.Query.phone.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.phone.tag,
            (namespaces.xep0077_in_band, "phone")
        )
        self.assertIsNone(
            ibr_xso.Query.phone.default,
        )

    def test_url(self):
        self.assertIsInstance(
            ibr_xso.Query.url.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.url.tag,
            (namespaces.xep0077_in_band, "url")
        )
        self.assertIsNone(
            ibr_xso.Query.url.default,
        )

    def test_date(self):
        self.assertIsInstance(
            ibr_xso.Query.date.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.date.tag,
            (namespaces.xep0077_in_band, "date")
        )
        self.assertIsNone(
            ibr_xso.Query.date.default,
        )

    def test_misc(self):
        self.assertIsInstance(
            ibr_xso.Query.misc.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.misc.tag,
            (namespaces.xep0077_in_band, "misc")
        )
        self.assertIsNone(
            ibr_xso.Query.misc.default,
        )

    def test_text(self):
        self.assertIsInstance(
            ibr_xso.Query.text.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.text.tag,
            (namespaces.xep0077_in_band, "text")
        )
        self.assertIsNone(
            ibr_xso.Query.text.default,
        )

    def test_key(self):
        self.assertIsInstance(
            ibr_xso.Query.key.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            ibr_xso.Query.key.tag,
            (namespaces.xep0077_in_band, "key")
        )
        self.assertIsNone(
            ibr_xso.Query.key.default,
        )

    def test_registered(self):
        self.assertIsInstance(
            ibr_xso.Query.registered.xq_descriptor,
            xso.ChildFlag,
        )
        self.assertEqual(
            ibr_xso.Query.registered.tag,
            (namespaces.xep0077_in_band, "registered")
        )

    def test_remove(self):
        self.assertIsInstance(
            ibr_xso.Query.remove.xq_descriptor,
            xso.ChildFlag,
        )
        self.assertEqual(
            ibr_xso.Query.remove.tag,
            (namespaces.xep0077_in_band, "remove")
        )



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.misc` -- Miscellaneous XSOs
##########################################

This subpackage bundles XSO definitions for several XEPs. They do not get their
own subpackage because they often only define one or two XSOs without any logic
involved. The XSOs are often intended for re-use by other protocols.


Out of Band Data (:xep:`66`)
============================

.. autoclass:: OOBExtension

.. attribute:: aioxmpp.Message.xep0066_oob


Delayed Delivery (:xep:`203`)
=============================

.. autoclass:: Delay()

.. attribute:: aioxmpp.Message.xep0203_delay

   A :class:`Delay` instance which indicates that the message has been
   delivered with delay.


Stanza Forwarding (:xep:`297`)
==============================

.. autoclass:: Forwarded()


Chat Markers (:xep:`333`)
=========================

.. autoclass:: ReceivedMarker

.. autoclass:: DisplayedMarker

.. autoclass:: AcknowledgedMarker

.. attribute:: aioxmpp.Message.xep0333_marker

Pre-Authenticated Roster Subcription (:xep:`379`)
=================================================

.. autoclass:: Preauth

.. attribute:: aioxmpp.Presence.xep0379_preauth

   The pre-auth element associate with a subscription request.

"""

from .delay import Delay  # NOQA
from .forwarding import Forwarded  # NOQA
from .oob import OOBExtension  # NOQA
from .markers import ReceivedMarker, DisplayedMarker, AcknowledgedMarker  # NOQA
from .pars import Preauth  # NOQA



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import copy
import logging
import random
import unittest
import unittest.mock

import aioxmpp
import aioxmpp.private_xml
import aioxmpp.xso

import aioxmpp.bookmarks

from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine
)

TEST_JID1 = aioxmpp.JID.fromstr("foo@bar.baz")
TEST_JID2 = aioxmpp.JID.fromstr("bar@bar.baz")


class PrivateXMLSimulator:

    def __init__(self, *, delay=0):
        self.stored = {}
        self.delay = 0

    @asyncio.coroutine
    def get_private_xml(self, xso):
        payload = copy.deepcopy(
            self.stored.setdefault(xso.TAG[0], copy.deepcopy(xso))
        )
        return aioxmpp.private_xml.Query(payload)

    @asyncio.coroutine
    def set_private_xml(self, xso):
        if self.delay == 0:
            self.stored[xso.TAG[0]] = copy.deepcopy(xso)
        else:
            self.delay -= 1


@aioxmpp.private_xml.Query.as_payload_class
class ExampleXSO(aioxmpp.xso.XSO):
    TAG = ("urn:example:foo", "example")

    text = aioxmpp.xso.Text()

    def __init__(self, text=""):
        self.text = text


class TestPrivateXMLSimulator(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.private_xml = PrivateXMLSimulator()

    def tearDown(self):
        del self.cc
        del self.private_xml

    def test_retrieve_store_and_retrieve(self):
        before = None
        after = None

        @asyncio.coroutine
        def test_private_xml():
            nonlocal before, after
            before = (
                yield from self.private_xml.get_private_xml(ExampleXSO())
            ).registered_payload
            yield from self.private_xml.set_private_xml(ExampleXSO("foo"))
            after = (
                yield from self.private_xml.get_private_xml(ExampleXSO())
            ).registered_payload

        run_coroutine(test_private_xml())

        self.assertIsInstance(before, ExampleXSO)
        self.assertEqual(before.text, "")

        self.assertIsInstance(after, ExampleXSO)
        self.assertEqual(after.text, "foo")

    def test_store_and_retrieve(self):
        @asyncio.coroutine
        def test_private_xml():
            yield from self.private_xml.set_private_xml(ExampleXSO("foo"))
            return (
                yield from self.private_xml.get_private_xml(ExampleXSO())
            ).registered_payload

        res = run_coroutine(test_private_xml())
        self.assertIsInstance(res, ExampleXSO)
        self.assertEqual(res.text, "foo")

    def test_store_delay(self):
        results = []
        self.private_xml.delay = 3

        @asyncio.coroutine
        def test_private_xml():
            for i in range(5):
                yield from self.private_xml.set_private_xml(ExampleXSO("foo"))
                results.append((
                    yield from self.private_xml.get_private_xml(ExampleXSO())
                ).registered_payload)
        run_coroutine(test_private_xml())

        self.assertEqual(results[0].text, "")
        self.assertEqual(results[1].text, "")
        self.assertEqual(results[2].text, "")
        self.assertEqual(results[3].text, "foo")
        self.assertEqual(results[4].text, "foo")


class TestBookmarkClient(unittest.TestCase):

    def test_is_service(self):
        self.assertTrue(issubclass(
            aioxmpp.bookmarks.BookmarkClient,
            aioxmpp.service.Service
        ))

    def test_after_private_xml(self):
        self.assertIn(
            aioxmpp.private_xml.PrivateXMLService,
            aioxmpp.bookmarks.BookmarkClient.ORDER_AFTER
        )

    def test__stream_established_is_decorated(self):
        self.assertTrue(
            aioxmpp.service.is_depsignal_handler(
                aioxmpp.Client,
                "on_stream_established",
                aioxmpp.bookmarks.BookmarkClient._stream_established,
                defer=True,
            )
        )

    def test__stream_established(self):
        with unittest.mock.patch.object(
                self.s,
                "sync",
                CoroutineMock()) as sync:
            run_coroutine(self.s._stream_established())
        self.assertEqual(len(sync.mock_calls), 1)

    def setUp(self):
        self.cc = make_connected_client()
        self.private_xml = PrivateXMLSimulator()
        self.s = aioxmpp.bookmarks.BookmarkClient(self.cc, dependencies={
            aioxmpp.private_xml.PrivateXMLService: self.private_xml
        })

    def tearDown(self):
        del self.cc
        del self.private_xml
        del self.s

    def connect_mocks(self):
        self.on_added = unittest.mock.Mock()
        self.on_added.return_value = None
        self.on_removed = unittest.mock.Mock()
        self.on_removed.return_value = None
        self.on_changed = unittest.mock.Mock()
        self.on_changed.return_value = None

        self.s.on_bookmark_added.connect(self.on_added)
        self.s.on_bookmark_removed.connect(self.on_removed)
        self.s.on_bookmark_changed.connect(self.on_changed)

    def test__get_bookmarks(self):
        with unittest.mock.patch.object(
                self.private_xml,
                "get_private_xml",
                new=CoroutineMock()) as get_private_xml_mock:
            get_private_xml_mock.return_value.registered_payload.bookmarks = \
                unittest.mock.sentinel.result
            res = run_coroutine(self.s._get_bookmarks())

        self.assertIs(res, unittest.mock.sentinel.result)
        self.assertEqual(
            len(get_private_xml_mock.mock_calls),
            1
        )
        (_, (arg,), kwargs), = get_private_xml_mock.mock_calls
        self.assertEqual(len(kwargs), 0)
        self.assertIsInstance(arg, aioxmpp.bookmarks.Storage)
        self.assertEqual(len(arg.bookmarks), 0)

    def test__set_bookmarks(self):
        bookmarks = []
        bookmarks.append(
            aioxmpp.bookmarks.Conference(
                "Coven", aioxmpp.JID.fromstr("coven@example.com")),
        )
        bookmarks.append(
            aioxmpp.bookmarks.URL(
                "Interesting",
                "http://example.com/"),
        )

        with unittest.mock.patch.object(
                self.private_xml,
                "set_private_xml",
                new=CoroutineMock()) as set_private_xml_mock:
            run_coroutine(self.s._set_bookmarks(bookmarks))

        self.assertEqual(
            len(set_private_xml_mock.mock_calls),
            1
        )
        (_, (arg,), kwargs), = set_private_xml_mock.mock_calls
        self.assertEqual(len(kwargs), 0)
        self.assertEqual(arg.bookmarks, bookmarks)

    def test__set_bookmarks_failure(self):
        bookmarks = unittest.mock.sentinel.something_else
        with unittest.mock.patch.object(
                self.private_xml,
                "set_private_xml",
                new=CoroutineMock()) as set_private_xml_mock:
            with self.assertRaisesRegex(
                    TypeError,
                    "can only assign an iterable$"):
                run_coroutine(self.s._set_bookmarks(bookmarks))

        self.assertEqual(
            len(set_private_xml_mock.mock_calls),
            0
        )

    def test_sync(self):
        on_added = unittest.mock.Mock()
        on_added.return_value = None
        on_removed = unittest.mock.Mock()
        on_removed.return_value = None
        on_changed = unittest.mock.Mock()
        on_changed.return_value = None

        self.s.on_bookmark_added.connect(on_added)
        self.s.on_bookmark_removed.connect(on_removed)
        self.s.on_bookmark_changed.connect(on_changed)

        with unittest.mock.patch.object(
                self.private_xml,
                "get_private_xml",
                new=CoroutineMock()) as get_private_xml_mock:
            result = aioxmpp.private_xml.Query(
                aioxmpp.bookmarks.Storage()
            )
            result.registered_payload.bookmarks.append(
                aioxmpp.bookmarks.Conference(
                    jid=aioxmpp.JID.fromstr("foo@bar.baz"),
                    name="foo",
                    nick="quux"
                )
            )
            get_private_xml_mock.return_value = result

            run_coroutine(self.s.sync())
            run_coroutine(self.s.sync())

            result = aioxmpp.private_xml.Query(
                aioxmpp.bookmarks.Storage()
            )
            result.registered_payload.bookmarks.append(
                aioxmpp.bookmarks.Conference(
                    jid=TEST_JID1,
                    name="foo",
                    nick="quux"
                )
            )
            result.registered_payload.bookmarks.append(
                aioxmpp.bookmarks.Conference(
                    jid=TEST_JID1,
                    name="foo",
                    nick="quuux"
                )
            )
            get_private_xml_mock.return_value = result

            run_coroutine(self.s.sync())
            run_coroutine(self.s.sync())

            result = aioxmpp.private_xml.Query(
                aioxmpp.bookmarks.Storage()
            )
            result.registered_payload.bookmarks.append(
                aioxmpp.bookmarks.Conference(
                    jid=TEST_JID1,
                    name="foo",
                    nick="quux"
                )
            )
            get_private_xml_mock.return_value = result

            run_coroutine(self.s.sync())
            run_coroutine(self.s.sync())

            result = aioxmpp.private_xml.Query(
                aioxmpp.bookmarks.Storage()
            )
            result.registered_payload.bookmarks.append(
                aioxmpp.bookmarks.Conference(
                    jid=aioxmpp.JID.fromstr("foo@bar.baz"),
                    name="foo",
                    nick="quuux"
                )
            )
            get_private_xml_mock.return_value = result

            run_coroutine(self.s.sync())
            run_coroutine(self.s.sync())

        self.assertEqual(
            len(on_added.mock_calls), 2
        )

        self.assertEqual(
            len(on_removed.mock_calls), 1
        )

        self.assertEqual(
            len(on_changed.mock_calls), 1
        )

    def test_set_bookmarks(self):
        bookmarks = [
            aioxmpp.bookmarks.URL("An URL", "http://foo.bar/"),
            aioxmpp.bookmarks.Conference(
                "Coven",
                aioxmpp.JID.fromstr("coven@conference.shakespeare.lit"),
                nick="Wayward Sister"
            )
        ]

        self.connect_mocks()
        run_coroutine(self.s.set_bookmarks(bookmarks))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 2)

    def test_add_bookmark(self):
        self.connect_mocks()
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))

        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.on_added.assert_called_once_with(bookmark)

    def test_add_bookmark_delay(self):
        self.private_xml.delay = 3
        self.connect_mocks()
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))

        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.on_added.assert_called_once_with(bookmark)

    def test_add_bookmark_delay_raises(self):
        self.private_xml.delay = 4
        with self.assertRaises(RuntimeError):
            with unittest.mock.patch.object(self.s, "_diff_emit_update") as f:
                bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
                run_coroutine(self.s.add_bookmark(bookmark))

        # check that _diff_emit_update is called
        self.assertEqual(len(f.mock_calls), 1)

    def test_add_bookmark_set_raises(self):
        class TokenException(Exception):
            pass

        def set_bookmarks(*args, **kwargs):
            raise TokenException

        with contextlib.ExitStack() as e:
            e.enter_context(self.assertRaises(TokenException))
            diff_emit_update = e.enter_context(
                unittest.mock.patch.object(self.s, "_diff_emit_update",)
            )
            e.enter_context(
                unittest.mock.patch.object(self.s, "_set_bookmarks",
                                           set_bookmarks)
            )
            bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
            run_coroutine(self.s.add_bookmark(bookmark))

        # check that _diff_emit_update is called
        self.assertEqual(len(diff_emit_update.mock_calls), 1)

    def test_add_bookmark_already_present(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))
        stored = run_coroutine(
            self.private_xml.get_private_xml(aioxmpp.bookmarks.Storage())
        )
        self.assertCountEqual(self.s._bookmark_cache,
                              [bookmark])
        self.assertCountEqual(stored.registered_payload.bookmarks,
                              [bookmark])
        self.connect_mocks()
        run_coroutine(self.s.add_bookmark(bookmark))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)

    def test_discard_bookmark(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))
        self.connect_mocks()
        run_coroutine(self.s.discard_bookmark(bookmark))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_removed.assert_called_once_with(bookmark)

    def test_discard_bookmark_delay(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))
        self.private_xml.delay = 3
        self.connect_mocks()
        run_coroutine(self.s.discard_bookmark(bookmark))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_removed.assert_called_once_with(bookmark)

    def test_discard_bookmark_delay_raises(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))
        self.private_xml.delay = 4
        with contextlib.ExitStack() as e:
            e.enter_context(self.assertRaises(RuntimeError))
            f = e.enter_context(
                unittest.mock.patch.object(self.s, "_diff_emit_update")
            )
            run_coroutine(self.s.discard_bookmark(bookmark))
        # check that _diff_emit_update is called
        self.assertEqual(len(f.mock_calls), 1)

    def test_discard_bookmark_set_raises(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))

        class TokenException(Exception):
            pass

        def set_bookmarks(*args, **kwargs):
            raise TokenException

        with contextlib.ExitStack() as e:
            e.enter_context(self.assertRaises(TokenException))
            diff_emit_update = e.enter_context(
                unittest.mock.patch.object(self.s, "_diff_emit_update",)
            )
            e.enter_context(
                unittest.mock.patch.object(self.s, "_set_bookmarks",
                                           set_bookmarks)
            )
            run_coroutine(self.s.discard_bookmark(bookmark))

        # check that _diff_emit_update is called
        self.assertEqual(len(diff_emit_update.mock_calls), 1)

    def test_discard_bookmark_removes_one(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.set_bookmarks([bookmark, bookmark]))
        self.connect_mocks()
        run_coroutine(self.s.discard_bookmark(bookmark))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_removed.assert_called_once_with(bookmark)
        self.assertCountEqual(self.s._bookmark_cache, [bookmark])

    def test_discard_bookmark_already_gone(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        self.connect_mocks()
        run_coroutine(self.s.discard_bookmark(bookmark))
        self.assertEqual(len(self.on_changed.mock_calls), 0)
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)

    def test_update_bookmark(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))

        self.connect_mocks()
        new_bookmark = copy.copy(bookmark)
        new_bookmark.name = "THE URL"
        run_coroutine(self.s.update_bookmark(bookmark, new_bookmark))
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_changed.assert_called_once_with(bookmark, new_bookmark)

    def test_update_bookmark_delay(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))
        self.private_xml.delay = 3
        self.connect_mocks()
        new_bookmark = copy.copy(bookmark)
        new_bookmark.name = "THE URL"
        run_coroutine(self.s.update_bookmark(bookmark, new_bookmark))
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_changed.assert_called_once_with(bookmark, new_bookmark)

    def test_update_bookmark_delay_raises(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        new_bookmark = copy.copy(bookmark)
        new_bookmark.name = "THE URL"
        run_coroutine(self.s.add_bookmark(bookmark))

        self.private_xml.delay = 4
        with contextlib.ExitStack() as e:
            e.enter_context(self.assertRaises(RuntimeError))
            f = e.enter_context(
                unittest.mock.patch.object(self.s, "_diff_emit_update")
            )
            run_coroutine(self.s.update_bookmark(bookmark, new_bookmark))
        # check that _diff_emit_update is called
        self.assertEqual(len(f.mock_calls), 1)

    def test_update_bookmark_set_raises(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        new_bookmark = copy.copy(bookmark)
        new_bookmark.name = "THE URL"
        run_coroutine(self.s.add_bookmark(bookmark))

        class TokenException(Exception):
            pass

        def set_bookmarks(*args, **kwargs):
            raise TokenException

        with contextlib.ExitStack() as e:
            e.enter_context(self.assertRaises(TokenException))
            diff_emit_update = e.enter_context(
                unittest.mock.patch.object(self.s, "_diff_emit_update",)
            )
            e.enter_context(
                unittest.mock.patch.object(self.s, "_set_bookmarks",
                                           set_bookmarks)
            )
            run_coroutine(self.s.update_bookmark(bookmark, new_bookmark))

        # check that _diff_emit_update is called
        self.assertEqual(len(diff_emit_update.mock_calls), 1)

    def test_concurrent_update_bookmark(self):
        bookmark = aioxmpp.bookmarks.URL("An URL", "http://foo.bar/")
        run_coroutine(self.s.add_bookmark(bookmark))

        self.private_xml.stored["storage:bookmarks"].bookmarks.clear()

        self.connect_mocks()
        new_bookmark = copy.copy(bookmark)
        new_bookmark.name = "THE URL"
        run_coroutine(self.s.update_bookmark(bookmark, new_bookmark))
        self.assertEqual(len(self.on_removed.mock_calls), 0)
        self.assertEqual(len(self.on_added.mock_calls), 0)
        self.on_changed.assert_called_once_with(bookmark, new_bookmark)

    def test_on_change_from_two_branches(self):
        pass

    def test_fuzz_bookmark_changes(self):
        bookmark_list = []

        logging.info(
            "This is a fuzzing test it may fail or not fail randomly"
            " depending on the chosen seed."
            "If it fails, please report a bug which includes "
            "the random generator state given in the next log message"
        )
        logging.info("The random seed is %s", random.getstate())

        def on_added(added):
            bookmark_list.append(added)

        def on_removed(removed):
            bookmark_list.remove(removed)

        def on_changed(old, new):
            bookmark_list.remove(old)
            bookmark_list.append(new)

        self.s.on_bookmark_added.connect(on_added)
        self.s.on_bookmark_removed.connect(on_removed)
        self.s.on_bookmark_changed.connect(on_changed)

        def random_nick():
            return "foo{}".format(random.randint(0, 5))

        def random_name():
            return "name{}".format(random.randint(0, 5))

        def random_pw():
            return "name{}".format(random.randint(0, 5))

        jids = [aioxmpp.JID.fromstr("foo{}@bar.baz".format(i))
                for i in range(5)]

        def random_jid():
            return random.choice(jids)

        def random_url():
            return "http://foo{}.bar/".format(random.randint(0, 5))

        for i in range(100):
            operation = random.randint(0, 100)
            if operation < 20:
                if random.randint(0, 1):
                    bookmark = aioxmpp.bookmarks.Conference(
                        random_name(),
                        random_jid(),
                        nick=random_nick(),
                        password=random_pw(),
                        autojoin=bool(random.randint(0, 1)),
                    )
                else:
                    bookmark = aioxmpp.bookmarks.URL(
                        random_name(),
                        random_url(),
                    )

                run_coroutine(self.s.add_bookmark(bookmark))
            elif operation < 30:
                if not bookmark_list:
                    continue

                run_coroutine(self.s.discard_bookmark(
                    bookmark_list[random.randrange(len(bookmark_list))]
                ))
            else:
                if not bookmark_list:
                    continue

                to_change = bookmark_list[random.randrange(len(bookmark_list))]
                changed = copy.copy(to_change)

                if type(to_change) is aioxmpp.bookmarks.Conference:
                    if random.randint(0, 4) == 0:
                        changed.name = random_name()
                    if random.randint(0, 4) == 0:
                        changed.jid = random_jid()
                    if random.randint(0, 4) == 0:
                        changed.nick = random_nick()
                    if random.randint(0, 4) == 0:
                        changed.password = random_pw()
                    if random.randint(0, 4) == 0:
                        changed.autojoin = bool(random.randint(0, 1))
                else:
                    if random.randint(0, 2) == 0:
                        changed.name = random_name()
                    if random.randint(0, 2) == 0:
                        changed.url = random_url()
                run_coroutine(self.s.update_bookmark(to_change, changed))

            self.assertCountEqual(bookmark_list, self.s._bookmark_cache)



########################################################################
# File name: nonza.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.nonza` --- Non-stanza stream-level XSOs (Nonzas)
###############################################################

This module contains XSO models for stream-level elements which are not
stanzas. Since :xep:`0360`, these are called "nonzas".

.. versionchanged:: 0.5

   Before version 0.5, this module was called :mod:`aioxmpp.stream_xsos`.

General XSOs
============

.. autoclass:: StreamError()

.. autoclass:: StreamFeatures()

StartTLS related XSOs
=====================

.. autoclass:: StartTLSXSO()

.. autoclass:: StartTLSFeature()

.. autoclass:: StartTLS()

.. autoclass:: StartTLSProceed()

.. autoclass:: StartTLSFailure()

SASL related XSOs
=================

.. autoclass:: SASLXSO

.. autoclass:: SASLAuth

.. autoclass:: SASLChallenge

.. autoclass:: SASLResponse

.. autoclass:: SASLFailure

.. autoclass:: SASLSuccess

.. autoclass:: SASLAbort

Stream management related XSOs
==============================

.. autoclass:: SMXSO()

.. autoclass:: SMRequest()

.. autoclass:: SMAcknowledgement()

.. autoclass:: SMEnable()

.. autoclass:: SMEnabled()

.. autoclass:: SMResume()

.. autoclass:: SMResumed()

"""
import itertools
import warnings

from . import xso, errors, stanza

from .utils import namespaces


class StreamError(xso.XSO):
    """
    XSO representing a stream error.

    .. attribute:: text

       The text content of the stream error.

    .. attribute:: condition

       The RFC 6120 stream error condition.

    """

    TAG = (namespaces.xmlstream, "error")

    DECLARE_NS = {}

    text = xso.ChildText(
        tag=(namespaces.streams, "text"),
        attr_policy=xso.UnknownAttrPolicy.DROP,
        default=None,
        declare_prefix=None
    )

    condition_obj = xso.Child(
        [member.xso_class for member in errors.StreamErrorCondition],
        required=True,
    )

    def __init__(self,
                 condition=errors.StreamErrorCondition.UNDEFINED_CONDITION,
                 text=None):
        super().__init__()
        if not isinstance(condition, errors.StreamErrorCondition):
            condition = errors.StreamErrorCondition(condition)
            warnings.warn(
                "as of aioxmpp 1.0, stream error conditions must be members "
                "of the aioxmpp.errors.StreamErrorCondition enumeration",
                DeprecationWarning,
                stacklevel=2,
            )

        self.condition_obj = condition.xso_class()
        self.text = text

    @property
    def condition(self):
        return self.condition_obj.enum_member

    @condition.setter
    def condition(self, value):
        if not isinstance(value, errors.StreamErrorCondition):
            value = errors.StreamErrorCondition(value)
            warnings.warn(
                "as of aioxmpp 1.0, stream error conditions must be members "
                "of the aioxmpp.errors.StreamErrorCondition enumeration",
                DeprecationWarning,
                stacklevel=2,
            )

        if self.condition == value:
            return
        self.condition_obj = value.xso_class()

    @classmethod
    def from_exception(cls, exc):
        instance = cls()
        instance.text = exc.text
        instance.condition = exc.condition
        return instance

    def to_exception(self):
        return errors.StreamError(
            condition=self.condition,
            text=self.text
        )


class StreamFeatures(xso.XSO):
    """
    XSO for collecting the supported stream features the remote advertises.

    To register a stream feature, use :meth:`register_child` with the
    :attr:`features` descriptor. A more fancy way to do the same thing is to
    use the :meth:`as_feature_class` classmethod as decorator for your feature
    XSO class.

    Adding new feature classes:

    .. automethod:: as_feature_class

    Querying features:

    .. method:: stream_features[FeatureClass]

       Obtain the first feature XSO which matches the `FeatureClass`. If no
       such XSO is contained in the :class:`StreamFeatures` instance
       `stream_features`, :class:`KeyError` is raised.

    .. method:: stream_features[FeatureClass] = feature

       Replace the stream features belonging to the given `FeatureClass` with
       the `feature` XSO.

       If the `FeatureClass` does not match the type of the `feature` XSO, a
       :class:`TypeError` is raised.

       It is legal to leave the FeatureClass out by specifying ``...``
       instead. In that case, the class is auto-detected from the `feature`
       object assigned.

    .. method:: del stream_features[FeatureClass]

       If any feature of the given `FeatureClass` type is in the
       `stream_features`, they are all removed.

       Otherwise, :class:`KeyError` is raised, to stay consistent with other
       mapping-like types.

    .. automethod:: get_feature

    .. automethod:: has_feature

    """

    TAG = (namespaces.xmlstream, "features")
    DECLARE_NS = {
        None: namespaces.xmlstream
    }

    # we drop unknown children
    UNKNOWN_CHILD_POLICY = xso.UnknownChildPolicy.DROP

    features = xso.ChildMap([])
    cruft = xso.Collector()

    @classmethod
    def as_feature_class(cls, other_cls):
        cls.register_child(cls.features, other_cls)
        return other_cls

    @classmethod
    def is_feature(cls, other_cls):
        return (cls.CHILD_MAP.get(other_cls.TAG, None) is
                cls.features.xq_descriptor)

    def __getitem__(self, feature_cls):
        tag = feature_cls.TAG
        try:
            return self.features[feature_cls.TAG][0]
        except IndexError:
            raise KeyError(feature_cls) from None

    def __setitem__(self, feature_cls, feature):
        if feature_cls is Ellipsis:
            feature_cls = type(feature)
        if not isinstance(feature, feature_cls):
            raise ValueError("incorrect XSO class supplied")
        self.features[feature_cls.TAG][:] = [feature]

    def __delitem__(self, feature_cls):
        items = self.features[feature_cls.TAG]
        if not items:
            raise KeyError(feature_cls)
        items.clear()

    def __contains__(self, other):
        raise TypeError("membership test not supported")

    def has_feature(self, feature_cls):
        """
        Return :data:`True` if the stream features contain a feature of the
        given `feature_cls` type. :data:`False` is returned otherwise.
        """
        return feature_cls.TAG in self.features

    def get_feature(self, feature_cls, default=None):
        """
        If a feature of the given `feature_cls` type is contained in the
        current stream features set, the first such instance is returned.

        Otherwise, `default` is returned.
        """
        try:
            return self[feature_cls]
        except KeyError:
            return default

    def __iter__(self):
        return itertools.chain(*self.features.values())


class StartTLSXSO(xso.XSO):
    """
    Base class for starttls related XSOs.

    This base class merely defines the namespaces to declare when serialising
    the derived XSOs
    """

    DECLARE_NS = {None: namespaces.starttls}


@StreamFeatures.as_feature_class
class StartTLSFeature(StartTLSXSO):
    """
    Start TLS capability stream feature
    """

    TAG = (namespaces.starttls, "starttls")

    class Required(xso.XSO):
        TAG = (namespaces.starttls, "required")

    required = xso.Child([Required], required=False)


class StartTLS(StartTLSXSO):
    """
    XSO indicating that the client wants to start TLS now.
    """

    TAG = (namespaces.starttls, "starttls")


class StartTLSFailure(StartTLSXSO):
    """
    Server refusing to start TLS.
    """
    TAG = (namespaces.starttls, "failure")


class StartTLSProceed(StartTLSXSO):
    """
    Server allows start TLS.
    """
    TAG = (namespaces.starttls, "proceed")


class SASLXSO(xso.XSO):
    DECLARE_NS = {
        None: namespaces.sasl
    }


class SASLAuth(SASLXSO):
    """
    Start SASL authentication.

    .. attribute:: mechanism

       The mechanism to authenticate with.

    .. attribute:: payload

       For mechanisms which use an initial client-supplied payload, this can be
       a string. It is automatically encoded as base64 according to the XMPP
       SASL specification.

    """

    TAG = (namespaces.sasl, "auth")

    mechanism = xso.Attr("mechanism")
    payload = xso.Text(
        type_=xso.Base64Binary(empty_as_equal=True),
        default=None
    )

    def __init__(self, mechanism, payload=None):
        super().__init__()
        self.mechanism = mechanism
        self.payload = payload


class SASLChallenge(SASLXSO):
    """
    A SASL challenge.

    .. attribute:: payload

       The (decoded) SASL payload as :class:`bytes`. Base64 en/decoding is
       handled by the XSO stack.

    """

    TAG = (namespaces.sasl, "challenge")

    payload = xso.Text(
        type_=xso.Base64Binary(empty_as_equal=True),
    )

    def __init__(self, payload):
        super().__init__()
        self.payload = payload


class SASLResponse(SASLXSO):
    """
    A SASL response.

    .. attribute:: payload

       The (decoded) SASL payload as :class:`bytes`. Base64 en/decoding is
       handled by the XSO stack.

    """

    TAG = (namespaces.sasl, "response")

    payload = xso.Text(
        type_=xso.Base64Binary(empty_as_equal=True)
    )

    def __init__(self, payload):
        super().__init__()
        self.payload = payload


class SASLFailure(SASLXSO):
    """
    Indication of SASL failure.

    .. attribute:: condition

       The condition which caused the authentication to fail.

    .. attribute:: text

       Optional human-readable text.

    """

    TAG = (namespaces.sasl, "failure")

    condition = xso.ChildTag(
        tags=[
            "aborted",
            "account-disabled",
            "credentials-expired",
            "encryption-required",
            "incorrect-encoding",
            "invalid-authzid",
            "invalid-mechanism",
            "malformed-request",
            "mechanism-too-weak",
            "not-authorized",
            "temporary-auth-failure",
        ],
        default_ns=namespaces.sasl,
        allow_none=False,
        declare_prefix=None,
    )

    text = xso.ChildText(
        (namespaces.sasl, "text"),
        attr_policy=xso.UnknownAttrPolicy.DROP,
        default=None,
        declare_prefix=None
    )

    def __init__(self, condition=(namespaces.sasl, "temporary-auth-failure")):
        super().__init__()
        self.condition = condition


class SASLSuccess(SASLXSO):
    """
    Indication of SASL success, with optional final payload supplied by the
    server.

    .. attribute:: payload

       The (decoded) SASL payload. Base64 en/decoding is handled by the XSO
       stack.

    """

    TAG = (namespaces.sasl, "success")

    payload = xso.Text(
        type_=xso.Base64Binary(empty_as_equal=True),
        default=None
    )


class SASLAbort(SASLXSO):
    """
    Request to abort the SASL authentication.
    """
    TAG = (namespaces.sasl, "abort")


class SMXSO(xso.XSO):
    """
    Base class for stream-management related XSOs.

    This base class merely defines the namespaces to declare when serializing
    the data.
    """

    DECLARE_NS = {
        None: namespaces.stream_management
    }


@StreamFeatures.as_feature_class
class StreamManagementFeature(SMXSO):
    """
    Stream management stream feature
    """
    TAG = (namespaces.stream_management, "sm")

    class Required(xso.XSO):
        TAG = (namespaces.stream_management, "required")

    class Optional(xso.XSO):
        TAG = (namespaces.stream_management, "optional")

    required = xso.Child([Required])
    optional = xso.Child([Optional])


class SMRequest(SMXSO):
    """
    A request for an SM acknowledgement (see :class:`SMAcknowledgement`).
    """
    TAG = (namespaces.stream_management, "r")


class SMAcknowledgement(SMXSO):
    """
    Response to a :class:`SMRequest`.

    .. attribute:: counter

       The counter as received by the remote side.

    """

    TAG = (namespaces.stream_management, "a")

    counter = xso.Attr(
        "h",
        type_=xso.Integer()
    )

    def __init__(self, counter=0, **kwargs):
        super().__init__(**kwargs)
        self.counter = counter

    def __repr__(self):
        return "<{}.{} counter={} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.counter,
            id(self),
        )


class SMEnable(SMXSO):
    """
    Request to enable stream management.

    .. attribute:: resume

       Set this to :data:`True` to request the capability of resuming the
       stream later.

    .. attribute:: max_

        Maximum time, as integer in seconds, for which the stream should be
        resumable after the connection dropped. Only relevant if
        :attr:`resume` is true and may be overridden by the server.

        .. versionadded:: 0.9

    """

    TAG = (namespaces.stream_management, "enable")

    resume = xso.Attr(
        "resume",
        type_=xso.Bool(),
        default=False
    )

    max_ = xso.Attr(
        "max",
        type_=xso.Integer(),
        default=None,
        validate=xso.ValidateMode.ALWAYS,
        validator=xso.NumericRange(min_=0),
    )

    def __init__(self, resume=False, max_=None):
        super().__init__()
        self.resume = resume
        self.max_ = max_

    def __repr__(self):
        return "<{}.{} resume={} max={} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.resume,
            self.max_,
            id(self),
        )


class SMEnabled(SMXSO):
    """
    Response to a :class:`SMEnable` request.

    .. attribute:: resume

       If :data:`True`, the peer allows resumption of the stream.

    .. attribute:: id_

       The SM-ID of the stream. This is required to resume later.

    .. attribute:: location

       A hostname-port pair which defines to which host the client shall
       connect to resume the stream.

    .. attribute:: max_

        Maximum time, as integer in seconds, for which the stream will be
        resumable after the connection dropped. Only relevant if
        :attr:`resume` is true.

    """
    TAG = (namespaces.stream_management, "enabled")

    resume = xso.Attr(
        "resume",
        type_=xso.Bool(),
        default=False
    )

    id_ = xso.Attr("id", default=None)

    location = xso.Attr(
        "location",
        type_=xso.ConnectionLocation(),
        default=None
    )

    max_ = xso.Attr(
        "max",
        type_=xso.Integer(),
        default=None,
        validate=xso.ValidateMode.ALWAYS,
        validator=xso.NumericRange(min_=0),
    )

    def __init__(self,
                 resume=False,
                 id_=None,
                 location=None,
                 max_=None):
        super().__init__()
        self.resume = resume
        self.id_ = id_
        self.location = location
        self.max_ = max_

    def __repr__(self):
        return "<{}.{} resume={} id={!r} location={!r} max={} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.resume,
            self.id_,
            self.location,
            self.max_,
            id(self),
        )


class SMResume(SMXSO):
    """
    Request resumption of a previously interrupted SM stream.

    .. attribute:: counter

       Set this to the value of the local incoming stanza counter.

    .. attribute:: previd

       Set this to the SM-ID, as received in :class:`SMEnabled`.

    """

    TAG = (namespaces.stream_management, "resume")

    counter = xso.Attr(
        "h",
        type_=xso.Integer()
    )
    previd = xso.Attr("previd")

    def __init__(self, counter, previd):
        super().__init__()
        self.counter = counter
        self.previd = previd

    def __repr__(self):
        return "<{}.{} counter={} previd={!r} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.counter,
            self.previd,
            id(self),
        )


class SMResumed(SMXSO):
    """
    Notification that SM resumption was successful, in response to
    :class:`SMResume`.

    .. attribute:: counter

       The stanza counter of the remote side.

    .. attribute:: previd

       The SM-ID of the stream.

    """

    TAG = (namespaces.stream_management, "resumed")

    counter = xso.Attr(
        "h",
        type_=xso.Integer())
    previd = xso.Attr("previd")

    def __init__(self, counter, previd):
        super().__init__()
        self.counter = counter
        self.previd = previd

    def __repr__(self):
        return "<{}.{} counter={} previd={!r} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.counter,
            self.previd,
            id(self),
        )


class SMFailed(SMXSO):
    """
    Server response to :class:`SMEnable` or :class:`SMResume` if stream
    management fails.
    """
    TAG = (namespaces.stream_management, "failed")

    condition = xso.Child(
        [member.xso_class for member in errors.ErrorCondition],
        required=True,
    )

    def __init__(self,
                 condition=errors.ErrorCondition.UNDEFINED_CONDITION,
                 **kwargs):
        super().__init__(**kwargs)
        self.condition = condition.to_xso()

    def __repr__(self):
        return "<{}.{} condition={!r} at 0x{:x}>".format(
            type(self).__module__,
            type(self).__qualname__,
            self.condition,
            id(self),
        )



########################################################################
# File name: test_rfc3921.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.rfc3921 as rfc3921
import aioxmpp.stanza as stanza
import aioxmpp.nonza as nonza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_session_namespace(self):
        self.assertEqual(
            "urn:ietf:params:xml:ns:xmpp-session",
            namespaces.rfc3921_session
        )


class TestSession(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            rfc3921.Session,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            rfc3921.Session.TAG,
            (namespaces.rfc3921_session, "session")
        )

    def test_declare_ns(self):
        self.assertDictEqual(
            rfc3921.Session.DECLARE_NS,
            {
                None: namespaces.rfc3921_session
            }
        )

    def test_is_iq_payload(self):
        self.assertIn(
            rfc3921.Session.TAG,
            stanza.IQ.CHILD_MAP
        )
        self.assertIs(
            stanza.IQ.CHILD_MAP[rfc3921.Session.TAG],
            stanza.IQ.payload.xq_descriptor
        )

    def test_child_policy(self):
        self.assertEqual(
            rfc3921.Session.UNKNOWN_CHILD_POLICY,
            xso.UnknownChildPolicy.DROP
        )

    def test_attr_policy(self):
        self.assertEqual(
            rfc3921.Session.UNKNOWN_ATTR_POLICY,
            xso.UnknownAttrPolicy.DROP
        )


class TestSessionFeature(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            rfc3921.SessionFeature,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            rfc3921.SessionFeature.TAG,
            (namespaces.rfc3921_session, "session")
        )

    def test_is_stream_feature(self):
        self.assertTrue(nonza.StreamFeatures.is_feature(
            rfc3921.SessionFeature
        ))

    def test_child_policy(self):
        self.assertEqual(
            rfc3921.SessionFeature.UNKNOWN_CHILD_POLICY,
            xso.UnknownChildPolicy.DROP
        )

    def test_attr_policy(self):
        self.assertEqual(
            rfc3921.SessionFeature.UNKNOWN_ATTR_POLICY,
            xso.UnknownAttrPolicy.DROP
        )

    def test_optional(self):
        self.assertIsInstance(
            rfc3921.SessionFeature.optional,
            xso.ChildFlag
        )
        self.assertEqual(
            rfc3921.SessionFeature.optional.tag,
            (namespaces.rfc3921_session, "optional")
        )

# foo



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
from aioxmpp.e2etest import (  # NOQA
    setup_package as e2etest_setup_package,
    teardown_package,
)

import warnings


def setup_package():
    e2etest_setup_package()
    warnings.filterwarnings(
        "error",
        message=".+(Stream)?ErrorCondition",
        category=DeprecationWarning,
    )



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import multidict

import aioxmpp.stanza
import aioxmpp.xso

from aioxmpp.utils import namespaces

namespaces.xep0363_http_upload = "urn:xmpp:http:upload:0"


@aioxmpp.IQ.as_payload_class
class Request(aioxmpp.xso.XSO):
    """
    XSO to request an upload slot from the server.

    The parameters initialise the attributes below.

    .. attribute:: filename
        :annotation: : str

        The file name (without path, but possibly with "extension") of the
        file to upload. The server MAY use this in the URL.


    .. attribute:: size
        :annotation: : int

        The size of the file in bytes. This must be accurate and MUST also
        be used as ``Content-Length`` header in the PUT request.

    .. attribute:: content_type
        :annotation: : str

        The MIME type of the file. This MUST be set in the PUT request as
        ``Content-Type`` header.
    """

    TAG = namespaces.xep0363_http_upload, "request"

    filename = aioxmpp.xso.Attr("filename")

    size = aioxmpp.xso.Attr(
        "size",
        type_=aioxmpp.xso.Integer(),
    )

    content_type = aioxmpp.xso.Attr("content-type")

    def __init__(self, filename, size, content_type):
        super().__init__()
        self.filename = filename
        self.size = size
        self.content_type = content_type


class Header(aioxmpp.xso.XSO):
    TAG = namespaces.xep0363_http_upload, "header"

    name = aioxmpp.xso.Attr("name")

    value = aioxmpp.xso.Text()


class HeaderType(aioxmpp.xso.AbstractElementType):
    @staticmethod
    def get_xso_types():
        return (Header,)

    @classmethod
    def unpack(self, header_xso):
        return header_xso.name, header_xso.value

    @classmethod
    def pack(self, t):
        header_xso = Header()
        header_xso.name, header_xso.value = t
        return header_xso


class Put(aioxmpp.xso.XSO):
    """
    .. attribute:: url
        :annotation: : str

        The URL against which the PUT request must be made.

    .. attribute:: headers
        :annotation: : multidict.MultiDict

        The headers which MUST be used in the PUT request as
        :class:`multidict.MultiDict`, in addition to the ``Content-Type``
        and ``Content-Length`` headers.

        The headers are already sanitised according to :xep:`363` (see also
        :attr:`HEADER_WHITELIST`).

    .. attribute:: HEADER_WHITELIST

        This *class attribute* holds the list of headers which are allowed to
        be used by the server. This defaults to the list specified in
        :xep:`363`.

        .. warning::

            Changing the list of allowed headers may have unintended security
            implications.
    """

    HEADER_WHITELIST = (
        "Authorization",
        "Expires",
        "Cookie",
    )

    TAG = namespaces.xep0363_http_upload, "put"

    url = aioxmpp.xso.Attr("url")

    headers = aioxmpp.xso.ChildValueMultiMap(
        HeaderType
    )

    def xso_after_load(self):
        whitelist = self.HEADER_WHITELIST
        headers = list(self.headers.items())
        self.headers.clear()

        for key, value in headers:
            if key not in whitelist:
                continue
            value = value.replace("\n", "")
            self.headers.add(key, value)


class Get(aioxmpp.xso.XSO):
    """
    .. attribute:: url
        :annotation: : str

        The URL at which the file can be retrieved after uploading.
    """

    TAG = namespaces.xep0363_http_upload, "get"

    url = aioxmpp.xso.Attr("url")


@aioxmpp.IQ.as_payload_class
class Slot(aioxmpp.xso.XSO):
    """
    XSO representing the an upload slot provided by the server.

    .. attribute:: get

        Information about the GET request for the slot as :class:`.Get` XSO.

    .. attribute:: put

        Information about the PUT request for the slot as :class:`.Put` XSO.
    """

    TAG = namespaces.xep0363_http_upload, "slot"

    put = aioxmpp.xso.Child([Put])

    get = aioxmpp.xso.Child([Get])

    def validate(self):
        super().validate()

        if self.put is None:
            raise ValueError("missing PUT information")

        if self.get is None:
            raise ValueError("missing GET information")



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import collections
import itertools
import unittest
import unittest.mock

import aioxmpp.stanza as stanza
import aioxmpp.forms.xso as forms_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_form_namespace(self):
        self.assertEqual(
            "jabber:x:data",
            namespaces.xep0004_data
        )


class TestValue(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            forms_xso.Value,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "value"),
            forms_xso.Value.TAG
        )

    def test_value(self):
        self.assertIsInstance(
            forms_xso.Value.value,
            xso.Text
        )
        self.assertEqual(
            forms_xso.Value.value.default,
            ""
        )


class TestValueElement(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            forms_xso.ValueElement,
            xso.AbstractElementType
        ))

    def test_get_xso_types(self):
        t = forms_xso.ValueElement()
        self.assertCountEqual(
            t.get_xso_types(),
            [forms_xso.Value]
        )

    def test_unpack(self):
        t = forms_xso.ValueElement()
        v = forms_xso.Value()
        v.value = "foobar"
        self.assertEqual(
            v.value,
            t.unpack(v)
        )

    def test_pack(self):
        t = forms_xso.ValueElement()
        v = t.pack("foo")
        self.assertIsInstance(
            v,
            forms_xso.Value
        )
        self.assertEqual(v.value, "foo")


class TestOption(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            forms_xso.Option,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "option"),
            forms_xso.Option.TAG
        )

    def test_label_attr(self):
        self.assertIsInstance(
            forms_xso.Option.label,
            xso.Attr
        )
        self.assertEqual(
            (None, "label"),
            forms_xso.Option.label.tag
        )

    def test_value_attr(self):
        self.assertIsInstance(
            forms_xso.Option.value,
            xso.ChildText
        )
        self.assertEqual(
            (namespaces.xep0004_data, "value"),
            forms_xso.Option.value.tag
        )

    def test_reject_missing_value(self):
        opt = forms_xso.Option()
        with self.assertRaisesRegex(ValueError,
                                    "option is missing a value"):
            opt.validate()


class TestOptionElement(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            forms_xso.OptionElement,
            xso.AbstractElementType
        ))

    def test_get_xso_types(self):
        t = forms_xso.OptionElement()
        self.assertCountEqual(
            t.get_xso_types(),
            [forms_xso.Option]
        )

    def test_unpack(self):
        t = forms_xso.OptionElement()
        o = forms_xso.Option()
        o.label = "fnord"
        o.value = "foobar"
        self.assertEqual(
            ("foobar", "fnord"),
            t.unpack(o)
        )

    def test_pack(self):
        t = forms_xso.OptionElement()
        o = t.pack(("foo", "bar"))
        self.assertIsInstance(
            o,
            forms_xso.Option
        )
        self.assertEqual(o.value, "foo")
        self.assertEqual(o.label, "bar")


class TestFieldType(unittest.TestCase):
    def test_covers_all_types(self):
        self.assertSetEqual(
            {v.value for v in forms_xso.FieldType},
            {
                "boolean",
                "fixed",
                "hidden",
                "jid-multi",
                "jid-single",
                "list-multi",
                "list-single",
                "text-multi",
                "text-private",
                "text-single",
            }
        )

    def test_has_options(self):
        positive = [
            forms_xso.FieldType.LIST_MULTI,
            forms_xso.FieldType.LIST_SINGLE,
        ]

        for enum_value in forms_xso.FieldType:
            self.assertEqual(
                enum_value in positive,
                enum_value.has_options
            )

    def test_is_multi_valued(self):
        positive = [
            forms_xso.FieldType.LIST_MULTI,
            forms_xso.FieldType.TEXT_MULTI,
            forms_xso.FieldType.JID_MULTI,
        ]

        for enum_value in forms_xso.FieldType:
            self.assertEqual(
                enum_value in positive,
                enum_value.is_multivalued,
            )

    def test_allow_upcast(self):
        allowed = {
            (forms_xso.FieldType.TEXT_SINGLE,
             forms_xso.FieldType.TEXT_PRIVATE)
        }

        for t1, t2 in itertools.product(forms_xso.FieldType,
                                        forms_xso.FieldType):
            self.assertEqual(
                t1 == t2 or (t1, t2) in allowed,
                t1.allow_upcast(t2)
            )


class TestField(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            forms_xso.Field,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "field"),
            forms_xso.Field.TAG
        )

    def test_required_attr(self):
        self.assertIsInstance(
            forms_xso.Field.required,
            xso.ChildFlag
        )
        self.assertEqual(
            (namespaces.xep0004_data, "required"),
            forms_xso.Field.required.tag
        )

    def test_desc_attr(self):
        self.assertIsInstance(
            forms_xso.Field.desc,
            xso.ChildText
        )
        self.assertEqual(
            (namespaces.xep0004_data, "desc"),
            forms_xso.Field.desc.tag
        )

    def test_values_attr(self):
        self.assertIsInstance(
            forms_xso.Field.values,
            xso.ChildValueList
        )
        self.assertSetEqual(
            {
                forms_xso.Value
            },
            set(forms_xso.Field.values._classes)
        )
        self.assertIsInstance(
            forms_xso.Field.values.type_,
            forms_xso.ValueElement
        )

    def test_options_attr(self):
        self.assertIsInstance(
            forms_xso.Field.options,
            xso.ChildValueMap
        )
        self.assertIsInstance(
            forms_xso.Field.options.type_,
            forms_xso.OptionElement
        )
        self.assertSetEqual(
            {
                forms_xso.Option
            },
            set(forms_xso.Field.options._classes)
        )
        self.assertEqual(
            forms_xso.Field.options.mapping_type,
            collections.OrderedDict,
        )

    def test_var_attr(self):
        self.assertIsInstance(
            forms_xso.Field.var,
            xso.Attr
        )
        self.assertEqual(
            (None, "var"),
            forms_xso.Field.var.tag
        )

    def test_type_attr(self):
        self.assertIsInstance(
            forms_xso.Field.type_,
            xso.Attr
        )
        self.assertEqual(
            (None, "type"),
            forms_xso.Field.type_.tag
        )
        self.assertIsInstance(
            forms_xso.Field.type_.type_,
            xso.EnumCDataType
        )
        self.assertIs(
            forms_xso.Field.type_.type_.enum_class,
            forms_xso.FieldType,
        )
        self.assertIsNone(forms_xso.Field.type_.validator)
        self.assertEqual(
            forms_xso.Field.type_.default,
            None
        )

    def test_label_attr(self):
        self.assertIsInstance(
            forms_xso.Field.label,
            xso.Attr
        )
        self.assertEqual(
            (None, "label"),
            forms_xso.Field.label.tag
        )

    def test_reject_missing_var_for_non_fixed_fields(self):
        types = set(forms_xso.FieldType)
        types.discard(forms_xso.FieldType.FIXED)

        for type_ in types:
            f = forms_xso.Field()
            f.type_ = type_
            with self.assertRaisesRegex(ValueError,
                                        "missing attribute var"):
                f.validate()

    def test_accept_missing_var_for_fixed_fields(self):
        f = forms_xso.Field()
        f.type_ = forms_xso.FieldType.FIXED
        f.validate()

        f.var = "foobar"
        f.validate()

    def test_reject_options_for_non_list_fields(self):
        types = set(forms_xso.FieldType)
        types.discard(forms_xso.FieldType.LIST_MULTI)
        types.discard(forms_xso.FieldType.LIST_SINGLE)

        for type_ in types:
            f = forms_xso.Field()
            f.type_ = type_
            f.var = "foobar"
            f.options["foo"] = "bar"

            with self.assertRaisesRegex(ValueError,
                                        "unexpected option on non-list field"):
                f.validate()

    def test_accept_options_for_list_fields(self):
        types = {
            forms_xso.FieldType.LIST_SINGLE,
            forms_xso.FieldType.LIST_MULTI,
        }

        option = forms_xso.Option()
        option.value = "foobar"

        for type_ in types:
            f = forms_xso.Field()
            f.type_ = type_
            f.var = "foobar"
            f.options["foo"] = "bar"

            f.validate()

    def test_accept_options_for_None(self):
        option = forms_xso.Option()
        option.value = "foobar"

        f = forms_xso.Field()
        f.type_ = None
        f.var = "foobar"
        f.options["foo"] = "bar"

        f.validate()

    def test_reject_multiple_values_for_non_multi_fields(self):
        types = set(forms_xso.FieldType)
        types.discard(forms_xso.FieldType.LIST_MULTI)
        types.discard(forms_xso.FieldType.TEXT_MULTI)
        types.discard(forms_xso.FieldType.JID_MULTI)

        value = forms_xso.Value()

        for type_ in types:
            f = forms_xso.Field()
            f.type_ = type_
            f.var = "foobar"
            f.values.append(value)
            f.values.append(value)

            with self.assertRaisesRegex(ValueError,
                                        "too many values on non-multi field"):
                f.validate()

    def test_accept_multiple_values_on_multi_fields(self):
        types = {
            forms_xso.FieldType.LIST_MULTI,
            forms_xso.FieldType.TEXT_MULTI,
            forms_xso.FieldType.JID_MULTI,
        }

        value = forms_xso.Value()

        for type_ in types:
            f = forms_xso.Field()
            f.type_ = type_
            f.var = "foobar"
            f.values.append(value)
            f.values.append(value)

            f.validate()

    def test_accept_multiple_values_for_None(self):
        f = forms_xso.Field()
        f.type_ = None
        f.var = "foobar"
        f.values.append(forms_xso.Value())
        f.values.append(forms_xso.Value())

        f.validate()

    def test_reject_duplicate_option_values(self):
        f = forms_xso.Field()
        f.type_ = forms_xso.FieldType.LIST_MULTI
        f.var = "foobar"
        f.options["foo"] = "bar"
        f.options["baz"] = "bar"

        with self.assertRaisesRegex(
                ValueError,
                "duplicate option value"):
            f.validate()

    def test_init(self):
        f = forms_xso.Field()
        self.assertEqual(f.type_, forms_xso.FieldType.TEXT_SINGLE)
        self.assertSequenceEqual(f.values, [])
        self.assertDictEqual(f.options, {})
        self.assertFalse(f.required)
        self.assertIsNone(f.desc)
        self.assertIsNone(f.var)
        self.assertIsNone(f.label)

    def test_init_args(self):
        options = {
            "1": "Foo 1",
            "2": "Foo 2",
            "3": "Foo 3",
        }

        values = ["1", "2"]

        f = forms_xso.Field(
            type_=forms_xso.FieldType.LIST_MULTI,
            options=options,
            values=values,
            desc="description",
            required=True,
            label="fnord"
        )

        self.assertEqual(f.type_, forms_xso.FieldType.LIST_MULTI)

        self.assertDictEqual(f.options, options)
        self.assertIsNot(f.options, options)

        self.assertSequenceEqual(f.values, values)
        self.assertIsNot(f.values, values)

        self.assertEqual(f.desc, "description")

        self.assertIs(
            f.required,
            True,
        )

        self.assertEqual(f.label, "fnord")


class TestInstructions(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            forms_xso.Instructions,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "instructions"),
            forms_xso.Instructions.TAG
        )

    def test_value(self):
        self.assertIsInstance(
            forms_xso.Instructions.value,
            xso.Text,
        )
        self.assertEqual(
            forms_xso.Instructions.value.default,
            ""
        )


class TestInstructionsElement(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            forms_xso.InstructionsElement,
            xso.AbstractElementType
        ))

    def test_get_xso_types(self):
        t = forms_xso.InstructionsElement()
        self.assertCountEqual(
            t.get_xso_types(),
            [forms_xso.Instructions]
        )

    def test_unpack(self):
        t = forms_xso.InstructionsElement()
        v = forms_xso.Instructions()
        v.value = "foobar"
        self.assertEqual(
            v.value,
            t.unpack(v)
        )

    def test_pack(self):
        t = forms_xso.InstructionsElement()
        v = t.pack("foo")
        self.assertIsInstance(
            v,
            forms_xso.Instructions
        )
        self.assertEqual(v.value, "foo")


class TestAbstractItem(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            forms_xso.AbstractItem,
            xso.XSO
        ))

    def test_fields_attr(self):
        self.assertIsInstance(
            forms_xso.AbstractItem.fields,
            xso.ChildList
        )
        self.assertSetEqual(
            {
                forms_xso.Field,
            },
            set(forms_xso.AbstractItem.fields._classes)
        )

    def test_has_no_tag(self):
        self.assertFalse(hasattr(forms_xso.AbstractItem, "TAG"))


class TestItem(unittest.TestCase):
    def test_is_abstract_item(self):
        self.assertTrue(issubclass(
            forms_xso.Item,
            forms_xso.AbstractItem
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "item"),
            forms_xso.Item.TAG
        )


class TestReported(unittest.TestCase):
    def test_is_abstract_item(self):
        self.assertTrue(issubclass(
            forms_xso.Reported,
            forms_xso.AbstractItem
        ))

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "reported"),
            forms_xso.Reported.TAG
        )


class TestDataType(unittest.TestCase):
    def test_covers_all_types(self):
        self.assertSetEqual(
            {v.value for v in forms_xso.DataType},
            {
                "form",
                "submit",
                "cancel",
                "result",
            }
        )


class TestData(unittest.TestCase):
    def test_is_abstract_item(self):
        self.assertTrue(issubclass(
            forms_xso.Data,
            forms_xso.AbstractItem
        ))

    def test_init_requires_type(self):
        with self.assertRaises(TypeError):
            forms_xso.Data()

    def test_init(self):
        f = forms_xso.Data(type_=forms_xso.DataType.FORM)
        self.assertEqual(f.type_, forms_xso.DataType.FORM)

    def test_tag(self):
        self.assertEqual(
            (namespaces.xep0004_data, "x"),
            forms_xso.Data.TAG
        )

    def test_type_attr(self):
        self.assertIsInstance(
            forms_xso.Data.type_,
            xso.Attr
        )
        self.assertEqual(
            (None, "type"),
            forms_xso.Data.type_.tag
        )
        self.assertIsInstance(
            forms_xso.Data.type_.type_,
            xso.EnumCDataType
        )
        self.assertEqual(
            forms_xso.Data.type_.type_.enum_class,
            forms_xso.DataType,
        )

    def test_title_attr(self):
        self.assertIsInstance(
            forms_xso.Data.title,
            xso.ChildText
        )
        self.assertEqual(
            forms_xso.Data.title.tag,
            (namespaces.xep0004_data, "title"),
        )
        self.assertIsNone(
            forms_xso.Data.title.default,
        )

    def test_fields_attr(self):
        self.assertIsInstance(
            forms_xso.Data.fields,
            xso.ChildList
        )
        self.assertSetEqual(
            {
                forms_xso.Field,
            },
            set(forms_xso.Data.fields._classes)
        )

    def test_instructions_attr(self):
        self.assertIsInstance(
            forms_xso.Data.instructions,
            xso.ChildValueList
        )
        self.assertSetEqual(
            {
                forms_xso.Instructions
            },
            set(forms_xso.Data.instructions._classes)
        )
        self.assertIsInstance(
            forms_xso.Data.instructions.type_,
            forms_xso.InstructionsElement,
        )

    def test_items_attr(self):
        self.assertIsInstance(
            forms_xso.Data.items,
            xso.ChildList
        )
        self.assertSetEqual(
            {
                forms_xso.Item,
            },
            set(forms_xso.Data.items._classes)
        )

    def test_reported_attr(self):
        self.assertIsInstance(
            forms_xso.Data.reported,
            xso.Child
        )
        self.assertSetEqual(
            {
                forms_xso.Reported,
            },
            set(forms_xso.Data.reported._classes)
        )

    def test_validate_rejects_reported_or_items_if_type_is_not_result(self):
        types = [
            forms_xso.DataType.FORM,
            forms_xso.DataType.SUBMIT,
            forms_xso.DataType.CANCEL,
        ]

        rep = forms_xso.Reported()
        item = forms_xso.Item()

        for type_ in types:
            obj = forms_xso.Data(type_=type_)
            obj.reported = rep
            with self.assertRaisesRegex(ValueError, "report in non-result"):
                obj.validate()
            obj.reported = None
            obj.items.append(item)
            with self.assertRaisesRegex(ValueError, "report in non-result"):
                obj.validate()
            obj.items.clear()

    def test_validate_rejects_fields_for_results_if_report(self):
        field = forms_xso.Field()
        field.type_ = forms_xso.FieldType.FIXED
        obj = forms_xso.Data(type_=forms_xso.DataType.RESULT)
        obj.fields.append(field)
        obj.reported = forms_xso.Reported()

        with self.assertRaisesRegex(ValueError, "field in report result"):
            obj.validate()

        obj.reported = None
        obj.items.append(forms_xso.Item())

        with self.assertRaisesRegex(ValueError, "field in report result"):
            obj.validate()

    def test_validate_accepts_fields_for_results_without_report(self):
        field = forms_xso.Field()
        field.type_ = forms_xso.FieldType.FIXED
        obj = forms_xso.Data(type_=forms_xso.DataType.RESULT)
        obj.fields.append(field)

        obj.validate()

    def test_validate_reject_empty_reported(self):
        obj = forms_xso.Data(forms_xso.DataType.RESULT)
        obj.reported = forms_xso.Reported()

        with self.assertRaisesRegex(ValueError, "empty report header"):
            obj.validate()

    def test_validate_reject_empty_items(self):
        f = forms_xso.Field()
        f.var = "foobar"

        obj = forms_xso.Data(forms_xso.DataType.RESULT)
        obj.reported = forms_xso.Reported()
        obj.reported.fields.append(f)

        obj.items.append(forms_xso.Item())

        with self.assertRaisesRegex(
                ValueError,
                "field mismatch between row and header"):
            obj.validate()

    def test_validate_reject_mismatching_items(self):
        f = forms_xso.Field()
        f.var = "foobar"

        obj = forms_xso.Data(forms_xso.DataType.RESULT)
        obj.reported = forms_xso.Reported()
        obj.reported.fields.append(f)

        f2 = forms_xso.Field()
        f2.var = "fnord"

        item = forms_xso.Item()
        item.fields.append(f2)
        obj.items.append(item)

        with self.assertRaisesRegex(
                ValueError,
                "field mismatch between row and header"):
            obj.validate()

    def test_data_attribute_on_Message(self):
        self.assertIsInstance(
            stanza.Message.xep0004_data,
            xso.ChildList,
        )
        self.assertSetEqual(
            stanza.Message.xep0004_data._classes,
            {
                forms_xso.Data,
            }
        )

    def test_get_form_type(self):
        d = forms_xso.Data(type_=forms_xso.DataType.FORM)
        d.fields.append(
            forms_xso.Field(),
        )
        d.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.HIDDEN,
                var="FORM_TYPE",
                values=[unittest.mock.sentinel.form_type]
            ),
        )
        d.fields.append(
            forms_xso.Field(),
        )

        self.assertEqual(
            d.get_form_type(),
            unittest.mock.sentinel.form_type,
        )

    def test_get_form_type_returns_none_without_FORM_TYPE(self):
        d = forms_xso.Data(type_=forms_xso.DataType.FORM)
        d.fields.append(
            forms_xso.Field(),
        )
        d.fields.append(
            forms_xso.Field(),
        )

        self.assertIsNone(
            d.get_form_type(),
        )

    def test_get_form_type_detects_incorrect_FORM_TYPE(self):
        d = forms_xso.Data(type_=forms_xso.DataType.FORM)
        d.fields.append(
            forms_xso.Field(),
        )
        d.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.TEXT_SINGLE,
                var="FORM_TYPE",
                values=[unittest.mock.sentinel.form_type]
            ),
        )
        d.fields.append(
            forms_xso.Field(),
        )

        self.assertIsNone(
            d.get_form_type(),
        )

    def test_get_form_type_copes_with_malformed_FORM_TYPE(self):
        d = forms_xso.Data(type_=forms_xso.DataType.FORM)
        d.fields.append(
            forms_xso.Field(),
        )
        d.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.HIDDEN,
                var="FORM_TYPE",
                values=[]
            ),
        )
        d.fields.append(
            forms_xso.Field(),
        )

        self.assertIsNone(
            d.get_form_type(),
        )

    def test_get_form_type_copes_with_too_many_values(self):
        d = forms_xso.Data(type_=forms_xso.DataType.FORM)
        d.fields.append(
            forms_xso.Field(),
        )
        d.fields.append(
            forms_xso.Field(
                type_=forms_xso.FieldType.HIDDEN,
                var="FORM_TYPE",
                values=["foo", "bar"]
            ),
        )
        d.fields.append(
            forms_xso.Field(),
        )

        self.assertIsNone(
            d.get_form_type(),
        )
